import sys
import os
import inspect
import runpy
import argparse

from pygears import PluginBase, __version__, reg
from pygears.conf import inject, Inject
from pygears.conf.custom_settings import load_rc


@inject
def main(argv=sys.argv, config=Inject('entry')):

    args = config['parser'].parse_args(argv[1:])

    cmd = args.command


    kwds = vars(args)
    print(kwds)

    del kwds['command']

    func = config['cmds'][cmd]['entry']

    func(**kwds)


@inject
def cmd_register(
    path,
    cmd_entry,
    aliases=(),
    structural=False,
    tree=Inject('entry'),
    derived=False,
    help=None,
    dest=None):

    for i, p in enumerate(path[:-1]):
        cmds = tree['cmds']
        if p not in cmds:
            raise Exception(f'Parent command does not exist: {"/".join(path[:i])}')

        tree = tree['cmds'][p]

    name = path[-1]
    cmd_config = tree['cmds'].subreg(name)

    parents = []
    if derived:
        parents = [tree['baseparser']]

    cmd_config['parser'] = tree['subparsers'].add_parser(
        name, aliases=aliases, parents=parents, help=help)

    if structural:
        cmd_config['subparsers'] = cmd_config['parser'].add_subparsers(
            title='subcommands',
            help='subcommand help',
            dest=dest)

        import sys
        if sys.version_info[1] >= 7:
            cmd_config['subparsers'].required = cmd_entry is None

    cmd_config['baseparser'] = argparse.ArgumentParser(add_help=False)

    cmd_config.subreg('cmds')
    cmd_config['entry'] = cmd_entry

    return cmd_config


def entry(path, aliases=None):
    def wrapper(f):
        cmd_register(path, cmd_entry=f, aliases=aliases)

    return wrapper


def design_exec_entry(f):
    def cmd_entry(**kwds):
        if kwds['design'] is None:
            kwds['design'] = inspect.stack()[0][1]
        else:
            runpy.run_path(kwds['design'])

        kwds['design'] = os.path.abspath(os.path.expanduser(kwds['design']))

        load_rc('.pygears', os.path.dirname(kwds['design']))

        f(**kwds)

    return f


class EntryPlugin(PluginBase):
    @classmethod
    def bind(cls):
        parser = argparse.ArgumentParser(
            description='Hardware Design: A Functional Approach')
        parser.add_argument(
            '--version', action='version', version=f'pygears {__version__}')

        subparsers = parser.add_subparsers(title='subcommands', help='subcommand help', dest='command')

        reg.subreg('entry', {'parser': parser, 'subparsers': subparsers})
        reg['entry'].subreg('cmds')



from pygears import gear, IntfEmpty, module
from pygears.sim import clk
from pygears.typing import Maybe, Tuple, Uint, Union, Bool
from .demux import demux
from .ccat import ccat
from .shred import shred


@gear
def spy(din) -> b'din':
    pass


@gear
async def avail_spy(din) -> b'(Bool, din)':
    pass


def avail(din):
    dout0, dout1 = avail_spy(din)

    for p in din.consumers[:-1]:
        if p is dout1.producer:
            continue

        din.disconnect(p)
        dout1.connect(p)

    return dout0


@gear
async def sample(din, *, latency=1, hold=1, init=None) -> b"din":
    data = din.dtype() if init is None else din.dtype(init)
    valid = init is not None
    dout = module().dout

    while True:
        if latency == 0:
            try:
                data = din.get_nb()
                valid = True
            except IntfEmpty:
                pass

        if valid and dout.ready_nb():
            dout.put_nb(data)

            if not hold:
                valid = False

        if latency == 1:
            try:
                data = din.get_nb()
                valid = True
            except IntfEmpty:
                pass

        await clk()


@gear(hdl={'files': ['sample.sv']})
async def trigreg(din: Maybe['data'], *, latency=1, init=None) -> b'data':
    pass


@gear
def regmap(wr: Tuple[{'addr': Uint['w_addr'], 'data': 'data'}], *, addrmap, initmap={}, regtype={}):

    *reqs, dflt = ccat(wr['data'], wr['addr']) \
        | Union \
        | demux(mapping=addrmap)

    dflt | shred

    douts = []
    for i, r in enumerate(reqs):
        init = initmap.get(i, None)
        type_ = regtype.get(i, sample)
        if type_.func.__name__ == 'trigreg':
            r = r >> Maybe[Uint[len(r.dtype) - 1]]

        douts.append(type_(r, init=init))

    return tuple(douts)

    # rd_data = mux(rd, *tuple(spy(d) for d in douts)) \
    #     | union_collapse(t=wr.dtype[0])

    # return tuple(rd_data, *douts)



# Copyright 2017 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.
"""Control flow graph analysis.

Given a Python AST we construct a doubly linked control flow graph whose nodes
contain the AST of the statements. We can then perform forward analysis on this
CFG.

"""
from __future__ import absolute_import
import functools
import operator
import inspect

import ast as gast

from . import ir
from . import HDLVisitor
from .debug import hls_debug_log_enabled


def get_name(node):
    """Get the name of a variable.
    Args:
        node: A `Name`, `Subscript` or `Attribute` node.
    Returns:
        The name of the variable e.g. `'x'` for `x`, `x.i` and `x[i]`.
    """
    if isinstance(node, ir.Name):
        return node.name
    elif isinstance(node, (ir.SubscriptExpr, ir.AttrExpr)):
        return get_name(node.val)
    else:
        raise TypeError


def _get_target(target):
    if isinstance(target, (ir.Name, ir.SubscriptExpr, ir.AttrExpr)):
        return set([get_name(target)])
    elif isinstance(target, (ir.ConcatExpr, gast.List)):
        return set.union(*(_get_target(target) for target in target.operands))
    elif isinstance(target, (ir.Component)):
        # return set()
        return set([f'{target.val.name}.{target.field}'])
    else:
        return set()
        # breakpoint()
        # raise ValueError


def get_updated(node):
    """Return the variable names created or mutated by this statement.
    This function considers assign statements, augmented assign statements, and
    the targets of for loops, as well as function arguments.
    For example, `x[0] = 2` will return `x`, `x, y = 3, 4` will return `x` and
    `y`, `for i in range(x)` will return `i`, etc.
    Args:
        node: An AST node
    Returns:
        A set of variable names (strings) of all the variables created or mutated.
    """
    if isinstance(node, ir.AssignValue):
        return _get_target(node.target)
    elif isinstance(node, ir.RegReset):
        return set([node.target.name])
    else:
        return set()

    # if isinstance(node, gast.Assign):
    #     return set.union(*(_get_target(target) for target in node.targets))
    # elif isinstance(node, (gast.For, gast.AugAssign)):
    #     return _get_target(node.target)
    # elif isinstance(node, gast.AsyncWith):
    #     return set(i.optional_vars.id for i in node.items)
    # elif isinstance(node, gast.arguments):
    #     targets = set(arg.arg for arg in node.args + node.kwonlyargs)
    #     if node.vararg:
    #         targets.add(node.vararg.id)
    #     if node.kwarg:
    #         targets.add(node.kwarg.id)
    #     return targets
    # else:
    #     return set()


class Node:
    # """A node in the CFG."""
    # __slots__ = ['next', 'value', 'prev', 'sink', '_source']

    def __init__(self, value, next_=None, prev=None, source=None):
        if next_ is None:
            next_ = []

        if prev is None:
            prev = []

        self.next = next_
        self.prev = prev
        self.value = value
        self._source = None

        self.source = source
        self.sink = None

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, val):
        if val is not None:
            val.sink = self
        elif self._source is not None:
            self._source.sink = None

        self._source = val


def backlink(node):
    """Given a CFG with outgoing links, create incoming links."""
    seen = set()
    to_see = [node]
    while to_see:
        node = to_see.pop()
        seen.add(node)
        for succ in node.next:
            succ.prev.append(node)
            if succ not in seen:
                to_see.append(succ)


def forwardlink(node):
    """Given a CFG with outgoing links, create incoming links."""
    seen = set()
    to_see = [node]
    while to_see:
        node = to_see.pop()
        seen.add(node)
        for pred in node.prev:
            pred.next.append(node)
            if pred not in seen:
                to_see.append(pred)


class CFG(HDLVisitor):
    """Construct a control flow graph.

    Each statement is represented as a node. For control flow statements such
    as conditionals and loops the conditional itself is a node which either
    branches or cycles, respectively.

    Attributes:
        entry: The entry node, which contains the `gast.arguments` node of the
            function definition.
        exit: The exit node. This node is special because it has no value (i.e. no
            corresponding AST node). This is because Python functions can have
            multiple return statements.
    """
    def __init__(self):
        # The current leaves of the CFG
        self.head = []
        # A stack of continue statements
        self.continue_ = []
        # A stack of break nodes
        self.break_ = []

    def set_head(self, node):
        """Link this node to the current leaves."""
        for head in self.head:
            head.next.append(node)
        self.head[:] = []
        self.head.append(node)

    @classmethod
    def build_cfg(cls, node):
        """Build a CFG for a function.

    Args:
      node: A function definition the body of which to analyze.

    Returns:
      A CFG object.

    Raises:
      TypeError: If the input is not a function definition.
    """
        cfg = cls()
        cfg.head = []
        cfg.entry = cfg.visit(node)
        cfg.exit = cfg.entry.sink
        backlink(cfg.entry)
        return cfg

    def Module(self, block: ir.Module):
        node = self.BaseBlock(block)
        node.sink.value = ir.ModuleSink()
        return node

    def BaseBlock(self, block: ir.BaseBlock):
        node = Node(block)
        self.set_head(node)
        for stmt in block.stmts:
            self.visit(stmt)

        sink = Node(ir.BaseBlockSink(), source=node)
        self.set_head(sink)

        return node

    def Branch(self, block: ir.Branch):
        node = self.BaseBlock(block)
        node.sink.value = ir.BranchSink()

    def Statement(self, stmt: ir.Statement):
        node = Node(stmt)
        self.set_head(node)
        return node

    def LoopBlock(self, block: ir.LoopBlock):
        # Start a new level of nesting
        self.break_.append([])
        self.continue_.append([])
        # Handle the body
        node = self.BaseBlock(block)
        node.sink.value = ir.LoopBlockSink()

        self.head.extend(self.continue_.pop())
        self.set_head(node)

        # The break statements and the test go to the next node
        self.head.extend(self.break_.pop())

    def HDLBlock(self, block: ir.HDLBlock):
        branch_exits = []
        node = Node(block)

        for b in block.branches:
            self.set_head(node)
            self.visit(b)
            branch_exits.extend(self.head[:])
            self.head[:] = []

        self.head.extend(branch_exits)
        sink = Node(ir.HDLBlockSink(), source=node)
        self.set_head(sink)

        if not block.has_else:
            br_else_sink = Node(ir.BranchSink(), next_=[sink])
            br_else = Node(ir.Branch(), next_=[br_else_sink])
            br_else_sink.source = br_else
            node.next.append(br_else)

    def generic_visit(self, node):
        breakpoint()
        raise ValueError('unknown control flow')


class Forward(object):
    """Forward analysis on CFG.

  Args:
    label: A name for this analysis e.g. 'active' for activity analysis. The
        AST nodes in the CFG will be given annotations 'name_in', 'name_out',
        'name_gen' and 'name_kill' which contain the incoming values, outgoing
        values, values generated by the statement, and values deleted by the
        statement respectively.
    gen: A function which takes the CFG node as well as a set of incoming
        values. It must return a set of newly generated values by the statement
        as well as a set of deleted (killed) values.
    op: Either the AND or OR operator. If the AND operator is used it turns
        into forward must analysis (i.e. a value will only be carried forward
        if it appears on all incoming paths). The OR operator means that
        forward may analysis is done (i.e. the union of incoming values will be
        taken).
  """
    def __init__(self, label, gen, op=operator.or_):
        self.gen = gen
        self.op = op
        self.out_label = label + '_out'
        self.in_label = label + '_in'
        self.gen_label = label + '_gen'
        self.kill_label = label + '_kill'
        self.reaching = {}

    def visit(self, node):
        if node not in self.reaching:
            self.reaching[node] = {}

        reaching = self.reaching[node]

        if node.value:
            if 'out' in reaching:
                before = hash(reaching['out'])
            else:
                before = None

            preds = [
                self.reaching[pred]['out'] for pred in node.prev
                if 'out' in self.reaching.get(pred, {})
            ]
            if preds:
                incoming = functools.reduce(self.op, preds[1:], preds[0])
            else:
                incoming = frozenset()

            reaching['in'] = incoming
            gen, kill = self.gen(node, incoming)
            reaching['gen'] = gen
            reaching['kill'] = kill
            reaching['out'] = (incoming - kill) | gen

            if hash(reaching['out']) != before:
                for succ in node.next:
                    self.visit(succ)


def forward(node, analysis):
    """Perform a given analysis on all functions within an AST."""
    if not isinstance(analysis, Forward):
        raise TypeError('not a valid forward analysis object')

    cfg_obj = CFG.build_cfg(node)

    if hls_debug_log_enabled():
        draw_cfg(cfg_obj)

    analysis.visit(cfg_obj.entry)

    return node, cfg_obj.entry, analysis.reaching


def node_name(node):
    if isinstance(node.value, ir.BaseBlock):
        return f'{type(node.value)}'

    return str(node.value)


def dfs(node, G, visited, labels):
    if node in visited:
        return

    visited.add(node)
    labels[id(node)] = node_name(node)

    for n in node.next:
        G.add_edge(id(node), id(n))

        dfs(n, G, visited, labels)


def draw_cfg(cfg):
    import networkx as nx
    import matplotlib.pyplot as plt

    G = nx.DiGraph()
    visited = set()
    labels = {}

    dfs(cfg.entry, G, visited, labels)

    pos = nx.planar_layout(G)
    nx.draw(G, pos, font_size=16, with_labels=False)

    # for p in pos:  # raise text positions
    #     pos[p][1] += 0.07

    nx.draw_networkx_labels(G, pos, labels)
    plt.show()


class ReachingDefinitions(Forward):
    """Perform reaching definition analysis.

  Each statement is annotated with a set of (variable, definition) pairs.

  """
    def __init__(self, update=get_updated):
        def definition(node, incoming):
            if isinstance(node.value, ir.Await) and isinstance(node.value.expr, ir.Component):
                intf_name = node.value.expr.val.name
                # gen = frozenset([(f'{intf_name}.ready', node.value)])
                gen = frozenset()
                kill = frozenset(def_ for def_ in incoming if def_[0] == f'{intf_name}.data')
            else:
                definitions = update(node.value)
                gen = frozenset((id_, node) for id_ in definitions)
                kill = frozenset(def_ for def_ in incoming if def_[0] in definitions)

            return gen, kill

        super(ReachingDefinitions, self).__init__('definitions', definition)


class Defined(Forward):
    """Perform defined variable analysis.

  Each statement is annotated with a set of variables which are guaranteed to
  be defined at that point.
  """
    def __init__(self):
        def defined(node, incoming):
            gen = get_updated(node.value)
            return gen, frozenset()

        super(Defined, self).__init__('defined', defined, operator.and_)


class CfgDfs:
    def __init__(self):
        self.scopes = []

    def visit(self, node):
        for base_class in inspect.getmro(node.value.__class__):
            if hasattr(self, base_class.__name__):
                return getattr(self, base_class.__name__)(node)
        else:
            return self.generic_visit(node)

    def enter(self, node):
        for base_class in inspect.getmro(node.value.__class__):
            method_name = f'enter_{base_class.__name__}'
            if hasattr(self, method_name):
                return getattr(self, method_name)(node)

    def exit(self, node):
        for base_class in inspect.getmro(node.value.__class__):
            method_name = f'exit_{base_class.__name__}'
            if hasattr(self, method_name):
                return getattr(self, method_name)(node)

    @property
    def parent(self):
        return self.scopes[-1]

    def LoopBlock(self, node):
        skip = self.enter(node)
        if not skip:
            self.scopes.append(node)
            self.visit(node.next[0])
            self.scopes.pop()

        self.exit(node)

        return self.visit(node.next[1])

    def LoopBlockSink(self, node):
        return

    def BaseBlock(self, node):
        skip = self.enter(node)
        if not skip:
            self.scopes.append(node)
            self.visit(node.next[0])
            self.scopes.pop()

        self.exit(node)

        return self.generic_visit(node.sink)

    def HDLBlock(self, node):
        skip = self.enter(node)
        if not skip:
            self.scopes.append(node)
            for n in node.next:
                self.visit(n)

            self.scopes.pop()
        self.exit(node)

        return self.generic_visit(node.sink)

    def generic_visit(self, node):
        skip = self.enter(node)
        self.exit(node)
        if not skip:
            for n in node.next:
                self.visit(n)

    def HDLBlockSink(self, node: ir.HDLBlockSink):
        if node.source in self.scopes:
            return

        return self.generic_visit(node)

    def BaseBlockSink(self, node: ir.BaseBlockSink):
        if node.source in self.scopes:
            return

        return self.generic_visit(node)



from pygears.conf import PluginBase, reg

# from .bool import Bool
from .queue import Queue
from .array import Array
from .base import TemplateArgumentsError, typeof, Any, is_type, T
from .tuple import Tuple
from .uint import Int, Uint, Integer, Bool, Integral, code, decode
from .unit import Unit
from .union import Union, Maybe, some, Nothing
from .math import bitw, ceil_pow2, div, floor
from .tlm import TLM
from .flatten import flatten
from .expand import expand
from .factor import factor
from .fixp import Fixp, Ufixp, Fixpnumber
from .number import Number
from .float import Float
from .cast import cast, signed
from .saturate import saturate
from .qround import qround
from .match import get_match_conds, TypeMatchError, match
from .trunc import trunc

__all__ = [
    'Bool', 'Queue', 'TemplateArgumentsError', 'Tuple', 'Int', 'Uint', 'Unit', 'Union',
    'Maybe', 'Array', 'Float', 'bitw', 'div', 'floor', 'typeof', 'Any', 'TLM',
    'ceil_pow2', 'is_type', 'flatten', 'expand', 'factor', 'Ufixp', 'Fixp', 'Number',
    'Fixpnumber', 'Integral', 'cast', 'signed', 'code', 'decode',
    'saturate', 'qround', 'get_match_conds', 'TypeMatchError', 'match', 'trunc', 'T'
]


class TypingNamespacePlugin(PluginBase):
    @classmethod
    def bind(cls):
        reg['gear/type_arith'] = {}


class CoreTypesPlugin(TypingNamespacePlugin):
    @classmethod
    def bind(cls):
        reg['gear/type_arith/Union'] = Union
        reg['gear/type_arith/Tuple'] = Tuple
        reg['gear/type_arith/Uint'] = Uint
        reg['gear/type_arith/Float'] = Float
        reg['gear/type_arith/Int'] = Int
        reg['gear/type_arith/Integer'] = Integer
        reg['gear/type_arith/Integral'] = Integral
        reg['gear/type_arith/Unit'] = Unit
        reg['gear/type_arith/Fixp'] = Fixp
        reg['gear/type_arith/Ufixp'] = Ufixp
        reg['gear/type_arith/Bool'] = Bool
        reg['gear/type_arith/Queue'] = Queue
        reg['gear/type_arith/Array'] = Array
        reg['gear/type_arith/bitw'] = bitw
        reg['gear/type_arith/ceil_pow2'] = ceil_pow2
        reg['gear/type_arith/typeof'] = typeof
        reg['gear/type_arith/Any'] = Any
        reg['gear/type_arith/TLM'] = TLM
        reg['gear/type_arith/flatten'] = flatten
        reg['gear/type_arith/expand'] = expand
        reg['gear/type_arith/factor'] = factor
        reg['gear/type_arith/cast'] = cast



import asyncio

from .graph import get_consumer_tree, get_producer_queue, get_end_producer
from pygears import GearDone
from pygears.conf import PluginBase, reg, MultiAlternativeError
from pygears.core.port import InPort, OutPort, HDLConsumer, HDLProducer
from pygears.core.partial import Partial
from pygears.core.sim_event import SimEvent
from pygears.typing import TypeMatchError
from pygears.typing import typeof, Any
from pygears.typing.base import TypingMeta
from pygears.conf import inject, Inject
from .graph import get_sim_map_gear

gear_reg = {}


def operator_func_from_namespace(cls, name):
    def wrapper(self, *args, **kwargs):
        if name not in reg['gear/intf_oper']:
            raise Exception(f'Operator {name} is not supported.')

        operator_func = reg[f'gear/intf_oper/{name}']
        return operator_func(self, *args, **kwargs)

    return wrapper


def operator_methods_gen(cls):
    for name in cls.OPERATOR_SUPPORT:
        setattr(cls, name, operator_func_from_namespace(cls, name))
    return cls


@operator_methods_gen
class Intf:
    OPERATOR_SUPPORT = [
        '__getitem__', '__neg__', '__add__', '__and__', '__div__', '__eq__', '__floordiv__',
        '__ge__', '__gt__', '__invert__', '__le__', '__lt__', '__mod__', '__mul__', '__ne__',
        '__neg__', '__lshift__', '__rshift__', '__aiter__', '__sub__', '__xor__', '__truediv__',
        '__matmul__'
    ]

    def __init__(self, dtype):
        self.consumers = []
        self._end_consumers = None
        self._end_producer = None
        self.dtype = dtype
        self.producer = None
        self._in_queue = None
        self._out_queues = []
        self._done = False
        self._data = None

        self.events = {
            'put': SimEvent(),
            'put_out': SimEvent(),
            'ack': SimEvent(),
            'ack_in': SimEvent(),
            'pull_start': SimEvent(),
            'pull_done': SimEvent(),
            'finish': SimEvent()
        }

    @property
    def sole_intf(self):
        if self.producer:
            return len(self.producer.gear.out_ports) == 1
        else:
            return True

    @property
    def is_broadcast(self):
        return len(self.consumers) > 1

    @property
    def basename(self):
        producer_port = self.producer
        if isinstance(producer_port, HDLProducer):
            #TODO: Not really a producer port
            producer_port = self.consumers[0]

        port_name = producer_port.basename

        if isinstance(producer_port, InPort):
            return port_name
        elif ((not self.is_broadcast) and self.consumers
              and isinstance(self.consumers[0], OutPort)):
            return self.consumers[0].basename
        elif hasattr(self, 'var_name'):
            return self.var_name
        elif self.sole_intf:
            return f'{producer_port.gear.basename}'
        else:
            return f'{producer_port.gear.basename}_{port_name}'

    @property
    def parent(self):
        if isinstance(self.producer, InPort):
            return self.producer.gear
        elif isinstance(self.producer, OutPort):
            return self.producer.gear.parent
        elif len(self.consumers) == 1 and isinstance(self.consumers[0], OutPort):
            return self.consumers[0].gear

    @property
    def name(self):
        if not self.parent:
            return None

        return f'{self.parent.name}.{self.basename}'

    def __ior__(self, iout):
        if isinstance(iout, Partial):
            raise Exception(f"Output of the unresolved gear '{iout.func.__name__}' with"
                            f" arguments {iout.args} and parameters {iout.kwds},"
                            f" connected to '{self}': {str(MultiAlternativeError(iout.errors))}")
        elif not isinstance(iout, Intf):
            raise Exception(f"Cannot connect to {iout}")

        if iout.dtype != self.dtype:
            raise TypeError(f'Output interface of type "{repr(iout.dtype)}", cannot be connected '
                            f'to the interface of type "{repr(self.dtype)}"')

        iout.producer.consumer = self
        if self.producer is not None:
            raise Exception(
                f"Interface '{self}' is already connected to a producer '{self.producer.name}'\n")

        self.producer = iout.producer

        return self

    def __or__(self, other):
        if isinstance(other, Intf):
            raise Exception(f'Cannot connect interface {self} to the interface {other}\n'
                            f'Did you mean to connect to "{other.producer.gear.name}"?')

        if not (isinstance(other, (str, TypingMeta)) or (other is int) or (other is float)):
            return other.__ror__(self)

        operator_func = reg['gear/intf_oper/__or__']
        return operator_func(self, other)

    def source(self, port):
        self.producer = port
        port.consumer = self

    def disconnect(self, port):
        if port in self.consumers:
            self.consumers.remove(port)
            port.producer = None
        elif port == self.producer:
            port.consumer = None
            self.producer = None

    def connect(self, port):
        if self.consumers and isinstance(self.consumers[0], HDLConsumer):
            self.consumers.clear()

        self.consumers.append(port)
        port.producer = self

    def __repr__(self):
        return f'Intf({repr(self.dtype)})'

    def __str__(self):
        return f'Intf({self.dtype})'

    @property
    def end_consumers(self):
        if self._end_consumers is None:
            self._end_consumers = get_consumer_tree(self)

        return self._end_consumers

    @property
    def end_producer(self):
        if self._end_producer is None:
            self._end_producer = get_end_producer(self)

        return self._end_producer

    @property
    def in_queue(self):
        if self._in_queue is None:
            if self.producer is not None:
                # self._in_queue = self.producer.get_queue()
                self._in_queue = get_producer_queue(self)

        return self._in_queue

    @property
    def out_queues(self):
        if self._out_queues:
            return self._out_queues

        self._out_queues = [
            asyncio.Queue(maxsize=1, loop=reg['sim/simulator']) for _ in self.end_consumers
        ]

        for i, q in enumerate(self._out_queues):
            q.intf = self
            q.index = i

        return self._out_queues

    def put_nb(self, val):
        put_event = self.events['put']

        if self.dtype is not type(val):
            err = None
            try:
                if not typeof(self.dtype, Any):
                    val = self.dtype(val)
            except (TypeError, ValueError) as e:
                err = e

            if err:
                # TODO: when value cannot be represented, the error report can be terse
                raise TypeMatchError(f'{str(err)}\n, when converting output data "{repr(val)}"'
                                     f' from the "{reg["gear/current_module"].name}"'
                                     f' module to the type {repr(self.dtype)}')

        if put_event:
            put_event(self, val)

        for q, c in zip(self.out_queues, self.end_consumers):
            if c.consumer is None or c.consumer._done:
                raise GearDone

            put_event = c.consumer.events['put']
            if put_event:
                put_event(c.consumer, val)

            q.put_nowait(val)

    async def ready(self):
        gear_reg['current_sim'].phase = 'back'
        for q in self._out_queues:
            if q._unfinished_tasks:
                await q.join()

    def ready_nb(self):
        for q in self.out_queues:
            if q._unfinished_tasks:
                return False

        return True

    async def send(self, val):
        if reg['sim/simulator'].phase != 'forward':
            from pygears.sim import clk
            await clk()

        self.put_nb(val)

    async def put(self, val):
        self.put_nb(val)
        await self.ready()

    def empty(self):
        if self._data is not None:
            return False
        else:
            return self.in_queue.empty()

    def finish(self):
        '''Mark the interface as done, i.e. no more data will be transmitted.

        Informs also all consumer ports. If any task is waiting for this
        interface's data it is finishled. This is how the end of simulation
        propagates from the producers to the consumers.
        '''

        self.events['finish'](self)
        self._done = True

        for q, c in zip(self.out_queues, self.end_consumers):
            c.finish()
            for task in q._getters:
                task.cancel()

    @property
    def done(self):
        return self._done

    def pull_nb(self):
        if self._done:
            raise GearDone

        if self._data is None:
            self._data = self.in_queue.get_nowait()

        if isinstance(self._data, self.dtype):
            return self._data

        try:
            return self.dtype(self._data)
        except TypeError:
            return self._data

    def get_nb(self):
        val = self.pull_nb()
        self.ack()
        return val

    async def pull(self):
        ev = self.events
        e = ev['pull_start']
        if e:
            e(self)

        if self._done:
            raise GearDone

        val = self._data
        if val is None:
            gear_reg['current_sim'].phase = 'forward'
            val = await self.in_queue.get()

        e = ev['pull_done']
        if e:
            e(self)

        self._data = val
        if isinstance(val, self.dtype):
            return val

        try:
            return self.dtype(val)
        except TypeError:
            return val

    def ack(self):
        inq = self.in_queue
        if not inq._unfinished_tasks:
            return

        e = self.events['ack']
        if e:
            e(self)

        inq.task_done()
        if inq.intf.ready_nb():
            e = inq.intf.events['ack']
            e(inq.intf)

        self._data = None
        return

    async def get(self):
        val = await self.pull()
        self.ack()
        return val

    async def __aenter__(self):
        return await self.pull()

    async def __aexit__(self, exception_type, exception_value, traceback):
        if exception_type is None:
            self.ack()

    def __hash__(self):
        return id(self)


class IntfOperPlugin(PluginBase):
    @classmethod
    def bind(cls):
        reg['gear/intf_oper'] = {}
        global gear_reg
        gear_reg = reg['gear']._dict



# from pygears.registry import load_plugin_folder
# import os
# load_plugin_folder(os.path.join(os.path.dirname(__file__), 'modules'))

from . import log
from .sim import sim, artifacts_dir, sim_assert, timestep, clk, delta, sim_phase, SimFinish
from .sim import SimPlugin, schedule_to_finish, cosim, SimSetupDone, cosim_build_dir
from . import inst
from .extens.vcd import VCD
from .call import call


def verilate(top, *args, **kwds):
    cosim(top, 'verilator', *args, **kwds)


__all__ = [
    'sim', 'artifacts_dir', 'sim_assert', 'clk', 'delta', 'timestep',
    'log', 'sim_phase', 'schedule_to_finish', 'SimFinish', 'verilate',
    'SimPlugin', 'VCD', 'call', 'inst', 'SimSetupDone', 'cosim_build_dir'
]



import pytest
import pathlib
from functools import partial
import os


@pytest.hookimpl(hookwrapper=True)
def pytest_exception_interact(node, call, report):
    if not node.config.getoption('--pg'):
        yield
        return

    from pygears import reg
    from pygears.conf import TraceLevel
    if reg['trace/level'] == TraceLevel.debug:
        yield
        return

    import re
    ignore = reg['trace/ignore']
    r = re.compile('  File "(?P<fn>[^"]+)"')

    if hasattr(report.longrepr, 'reprtraceback'):
        for reprs in report.longrepr.reprtraceback.reprentries:
            lines = []
            for l in report.longrepr.reprtraceback.reprentries[0].lines:
                res = r.match(l)
                if res is not None:
                    fn = res.groupdict()['fn']
                    is_internal = any(fn.startswith(d) for d in ignore)
                    is_decorator_gen = '<decorator-gen' in fn
                    if is_internal or is_decorator_gen:
                        continue

                lines.append(l)

            reprs.lines = lines

    yield


@pytest.fixture(autouse=True)
def resdir(tmp_path_factory, pytestconfig):
    if not pytestconfig.getoption('--pg'):
        return None

    from pygears import clear, reg
    from pygears.conf.custom_settings import load_rc
    clear()
    load_rc('.pygears', pytestconfig.rootdir)

    resdir = pytestconfig.getoption("--pg-resdir")

    if resdir is None:
        resdir = str(tmp_path_factory.getbasetemp())

    from pygears import reg
    reg['results-dir'] = resdir

    return resdir


def hook_after(top, args, kwds, file_lock):
    file_lock.release()
    return False


def hook_before(top, args, kwds, config):
    import filelock
    from pygears import reg
    from pygears.sim import cosim_build_dir

    inst_id = cosim_build_dir(top)
    kwds['rebuild'] = False

    kwds['outdir'] = pathlib.Path(kwds['outdir'])
    if config.getoption('--pg-resdir') is None:
        # Xdist creates an extra level of folders. We want one compilation per
        # test run, so we need to go to the parent
        if 'PYTEST_XDIST_WORKER_COUNT' in os.environ:
            kwds['outdir'] = kwds['outdir'].parent

        kwds['outdir'] = kwds['outdir'] / 'sim'

    os.makedirs(kwds['outdir'], exist_ok=True)

    lock_fn = kwds['outdir'] / f'{inst_id}.lock'
    try:
        fl = filelock.FileLock(lock_fn, timeout=0)
        fl.acquire()
        reg['sim/hook/cosim_build_after'].append(partial(hook_after, file_lock=fl))
    except filelock.Timeout:
        with filelock.FileLock(lock_fn):
            pass

    return True


def pytest_runtest_call(item):
    if item.config.getoption('--pg'):
        from pygears import reg
        if item.config.getoption('--pg-reuse'):
            reg['sim/hook/cosim_build_before'].append(partial(hook_before, config=item.config))

    if item.config.getoption('--gearbox'):
        from gearbox.main import main_loop
        main_loop(str(pathlib.Path(item.module.__file__).parent), [], item.runtest, {})
        pytest.skip()


# If any of --pg switches is active, than --pg should be there too
def pytest_load_initial_conftests(args):
    if (any(a.startswith('--pg') for a in args) and not any(a == '--pg' for a in args)):
        args[:] = ['--pg'] + args

    if (not any(a.startswith('--tb') for a in args)):
        args[:] = ['--tb=native'] + args


# def pytest_sessionfinish(session, exitstatus):
#     breakpoint()
#     pass


def pytest_addoption(parser):
    parser.addoption(
        "--gearbox",
        action="store_true",
        help="Specify the result directory. Use /tmp by default.",
    )

    parser.addoption(
        "--pg",
        action="store_true",
        help="Use pygears pytest plugin",
    )

    parser.addoption(
        "--pg-resdir",
        type=str,
        default=None,
        help="Specify the result directory. Use /tmp by default.",
    )

    parser.addoption(
        "--pg-reuse",
        action="store_true",
        help="Specify the result directory. Use /tmp by default.",
    )



import functools
import os
from pygears import reg
from ...base_resolver import ResolverBase, ResolverTypeError
from ..svparse import parse
from pygears.util.fileio import find_in_dirs
from pygears.typing import code, is_type


class HDLFileResolver(ResolverBase):
    def __init__(self, node):
        self.node = node

        if self.impl_parse is None:
            raise ResolverTypeError

    @property
    def hdl_path_list(self):
        return reg[f'{self.lang}gen/include']

    @property
    def impl_path(self):
        return find_in_dirs(self.file_basename, self.hdl_path_list)

    @property
    @functools.lru_cache()
    def impl_parse(self):
        if self.impl_path:
            with open(self.impl_path, 'r') as f:
                return parse(f.read())

        return None

    @property
    def impl_params(self):
        return self.impl_parse[-1]

    @property
    def impl_intfs(self):
        parse_res = self.impl_parse()
        if parse_res:
            intfs = {}
            for i in parse_res[2]:
                intfs[i['name']] = i

            return intfs

        return None

    @property
    def params(self):
        if not self.impl_params:
            return {}

        params = {}
        for k, v in self.node.params.items():
            param_name = k.upper()
            param_valid_name = f'{param_name}_VALID'

            if (param_name in self.impl_params):
                if v is None:
                    if param_valid_name in self.impl_params:
                        params[param_valid_name] = 0
                    continue

                if is_type(v):
                    v = max(v.width, 1)

                err = None
                try:
                    v = code(v, int)
                except:
                    err = ValueError(
                        f'Cannot encode value "{v}" as integer, passed for HDL parameter "{param_name}"\n'
                        f' - when instantiating module "{self.node.name}"')

                if err:
                    raise err

                if (code(v, int) != int(self.impl_params[param_name]['val'])):
                    params[param_name] = code(v, int)

                if param_valid_name in self.impl_params:
                    params[param_valid_name] = 1

        return params

    @property
    def explicit(self):
        if 'hdl' not in self.node.meta_kwds:
            return False

        if 'impl' not in self.node.meta_kwds['hdl']:
            return False

        return self.node.meta_kwds['hdl']['impl']

    @property
    def file_basename(self):
        fn = self.node_def_name
        if 'hdl' in self.node.meta_kwds:
            if 'impl' in self.node.meta_kwds['hdl']:
                fn = self.node.meta_kwds['hdl']['impl']

        if not os.path.splitext(fn)[-1]:
            fn = f'{fn}.{self.lang}'

        return fn

    @property
    def files(self):
        files = [self.impl_path]
        if 'hdl' in self.node.meta_kwds:
            if 'files' in self.node.meta_kwds['hdl']:
                for fn in self.node.meta_kwds['hdl']['files']:
                    if not os.path.splitext(fn)[-1]:
                        fn = f'{fn}.{self.lang}'

                    files.append(fn)

        return files

    @property
    @functools.lru_cache()
    def module_name(self):
        return self.impl_parse[0]

    def generate(self, template_env, outdir):
        pass



from pygears import gear
from pygears.typing import Queue, Uint, Tuple


def chunk_type(dtypes, chunk_size):
    if (chunk_size == 1):
        return Queue[dtypes[0]]
    else:
        return Queue[Tuple[(dtypes[0][0], ) * chunk_size], 2]


@gear
def chunk_concat(cfg: Uint['Tn'],
                 *din: 'w_din{0}',
                 cnt_type=0,
                 chunk_size=1,
                 pad=0) -> b'chunk_type(din, chunk_size)':
    pass



from functools import singledispatch
from pygears.typing import typeof, Int, Fixp, Uint, Ufixp, is_type, Integer, Fixpnumber, Integral, trunc
from pygears.typing.uint import IntegerType, IntegralType
from pygears.typing.fixp import FixpnumberType
from .cast import cast
from . import code


@singledispatch
def type_saturate(cast_type, val):
    raise TypeError(f"Type '{repr(cast_type)}' unsupported, cannot saturate type '{val}' "
                    f"of type '{repr(type(val))}'")


@type_saturate.register(IntegerType)
def integral_type_saturate_resolver(cast_type, dtype):
    if dtype is not int and not typeof(dtype, Integer):
        raise TypeError(
            f"cannot saturate '{repr(dtype)}' to a different base type '{repr(cast_type)}'")

    return cast_type


@type_saturate.register(FixpnumberType)
def fixp_type_saturate_resolver(cast_type, dtype):
    if not typeof(dtype, Fixpnumber):
        raise TypeError(
            f"cannot saturate '{repr(dtype)}' to a different base type '{repr(cast_type)}'")

    if dtype.fract != cast_type.fract:
        raise TypeError(
            f"cannot saturate fixed point type '{repr(dtype)}' to a type '{repr(cast_type)}'"
            f" with a different fractional size ({dtype.fract} != {cast_type.fract})")

    return cast_type


@singledispatch
def value_saturate(cast_type, val, limits=None):
    raise ValueError(
        f"Saturating to type '{repr(cast_type)}' unsupported, cannot saturate value '{val}' "
        f"of type '{repr(type(val))}'")


@value_saturate.register(IntegralType)
def integral_saturate_resolver(t, data: Integral, limits=None):
    if not is_type(type(data)) and isinstance(data, int):
        conv_data = Integer(data)
    else:
        conv_data = data

    idin = code(data)

    if type(conv_data).signed == t.signed and type(conv_data).width <= t.width:
        if type(conv_data).signed:
            sign = code(data, int) >> (type(conv_data).width - 1)
            sign_exten = Uint[t.width - type(conv_data).width].max if sign else 0
            return t.decode((sign_exten << type(conv_data).width) | code(data, int))
        else:
            return code(conv_data, t)
    elif type(conv_data).signed and not t.signed:
        if idin[t.width:] == 0:
            return code(conv_data, t)
        elif conv_data < 0:
            return 0
        else:
            return t.max
    elif type(conv_data).signed and t.signed:
        # TODO: This 0 is not typecast, check why that happens
        if ((idin[t.width - 1:] == 0)
                or (idin[t.width - 1:] == Uint[type(conv_data).width - t.width + 1].max)):
            return code(conv_data, t)
        elif idin[-1]:
            return t.min
        else:
            return t.max
    else:
        if type(conv_data).width <= t.width or idin[t.width:] == 0:
            return code(conv_data, t)
        else:
            return t.max


def saturate(data, t, limits=None):
    if is_type(data):
        return type_saturate(t, data)
    else:
        return value_saturate(type_saturate(t, type(data)), data, limits)



from pygears.core.hier_node import HierVisitorBase

class HDLGearHierVisitor(HierVisitorBase):
    def RTLGear(self, node):
        gear = node.gear
        if hasattr(self, gear.definition.__name__):
            return getattr(self, gear.definition.__name__)(node)


def flow_visitor(cls):
    def svgen_action(top, conf):
        v = cls()
        v.conf = conf
        v.visit(top)
        return top

    return svgen_action


def is_gear_instance(node, definition):
    return node.gear.definition is definition



import importlib
import ctypes
import re
import os
import sys
import jinja2
from jinja2.ext import Extension
from pygears.typing import bitw, code, decode
from textwrap import dedent
from io import StringIO


def format_list(list_, pattern):
    return [pattern % s for s in list_]


def keymap(list_, key):
    return [item[key] for item in list_]


def isinput(list_):
    return [item for item in list_ if item["modport"] == "consumer"]


def isoutput(list_):
    return [item for item in list_ if item["modport"] == "producer"]


def startswith(field, s):
    return field.startswith(s)


def import_from(module, name):
    module = importlib.import_module(module)
    return getattr(module, name)


def import_(module):
    return importlib.import_module(module)


def debug(text):
    print(text)
    return ''


def get_port_config(modport, type_, name):
    return {
        'modport': modport,
        'name': name,
        'size': 1,
        'type': type_,
        'width': type_.width,
        'local_type': type_
    }


def get_port_intfs(node):
    intfs = []

    for p in node.in_ports:
        intfs.append(get_port_config('consumer', type_=p.dtype, name=p.basename))

    for p in node.out_ports:
        intfs.append(get_port_config('producer', type_=p.dtype, name=p.basename))

    return intfs

class TemplateEnv:
    def __init__(self, basedir):
        self.basedir = basedir
        self.templates = {}
        self.jenv = jinja2.Environment(
            extensions=['jinja2.ext.do', PythonExtension],
            trim_blocks=True,
            lstrip_blocks=True,
            undefined=jinja2.StrictUndefined)

        self.jenv.globals.update(
            zip=zip,
            len=len,
            int=int,
            max=max,
            tuple=tuple,
            isinstance=isinstance,
            bitw=bitw,
            enumerate=enumerate,
            import_from=import_from,
            import_=import_,
            code=code,
            decode=decode)

        self.jenv.filters['format_list'] = format_list
        self.jenv.filters['keymap'] = keymap
        self.jenv.filters['isinput'] = isinput
        self.jenv.filters['isoutput'] = isoutput
        self.jenv.filters['debug'] = debug
        self.jenv.tests['startswith'] = startswith

    # TODO: Revisit this function
    def port_intfs(self, node):
        return get_port_intfs(node)

    def load(self, tmplt_dir, tmplt_fn):
        key = os.path.join(self.basedir, tmplt_dir, tmplt_fn)
        if key not in self.templates:
            self.jenv.loader = jinja2.FileSystemLoader([self.basedir, tmplt_dir])
            template = self.jenv.get_template(tmplt_fn)
            self.templates[key] = template

        return self.templates[key]

    def render_string(self, string, context):
        return self.jenv.from_string(string).render(context)

    def render_local(self, fn, tmplt_fn, context):
        return self.render(os.path.dirname(fn), tmplt_fn, context)

    def render(self, tmplt_dir, tmplt_fn, context):
        return self.load(tmplt_dir, tmplt_fn).render(context)

var_name_regex = re.compile(r"l_(\d+)_(.+)")


class PythonExtension(Extension):
    # a set of names that trigger the extension.
    tags = {'py'}

    def __init__(self, environment: jinja2.Environment):
        super().__init__(environment)

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        body = parser.parse_statements(['name:endpy'], drop_needle=True)
        return jinja2.nodes.CallBlock(self.call_method('_exec_python',
                                                [jinja2.nodes.ContextReference(), jinja2.nodes.Const(lineno), jinja2.nodes.Const(parser.filename)]),
                               [], [], body).set_lineno(lineno)

    def _exec_python(self, ctx, lineno, filename, caller):
        # Remove access indentation
        code = dedent(caller())

        # Compile the code.
        compiled_code = compile("\n"*(lineno-1) + code, filename, "exec")

        # Create string io to capture stdio and replace it.
        sout = StringIO()
        stdout = sys.stdout
        sys.stdout = sout

        try:
            # Execute the code with the context parents as global and context vars and locals.
            exec(compiled_code, ctx.parent, ctx.vars)
        except Exception:
            raise
        finally:
            # Restore stdout whether the code crashed or not.
            sys.stdout = stdout

        # Get a set of all names in the code.
        code_names = set(compiled_code.co_names)

        # The the frame in the jinja generated python code.
        caller_frame = sys._getframe(2)

        # Loop through all the locals.
        for local_var_name in caller_frame.f_locals:
            # Look for variables matching the template variable regex.
            match = re.match(var_name_regex, local_var_name)
            if match:
                # Get the variable name.
                var_name = match.group(2)

                # If the variable's name appears in the code and is in the locals.
                if (var_name in code_names) and (var_name in ctx.vars):
                    # Copy the value to the frame's locals.
                    caller_frame.f_locals[local_var_name] = ctx.vars[var_name]
                    # Do some ctypes vodo to make sure the frame locals are actually updated.
                    ctx.exported_vars.add(var_name)
                    ctypes.pythonapi.PyFrame_LocalsToFast(
                        ctypes.py_object(caller_frame),
                        ctypes.c_int(1))

        # Return the captured text.
        return sout.getvalue()




import os
from pygears import find
from pygears.sim.extens.sim_extend import SimExtend
from pygears.sim.extens import graphviz


class SimTimelapse(SimExtend):
    node_state_color_map = {
        'forward': 'green',
        'back': 'orange',
        'finished': 'red'
    }

    edge_state_color_map = {
        'new': 'black',
        'forward': 'green',
        'back': 'orange',
        'finished': 'red'
    }

    def __init__(self, outdir='timelapse', dpi=60):
        super().__init__()
        self.outdir = os.path.abspath(
            os.path.expandvars(os.path.expanduser(outdir)))
        os.makedirs(self.outdir, exist_ok=True)
        self.g = graphviz.graph(find('/'), edge_fmt='{prod_gear} -> {cons_gear}')
        self.g.set_dpi(dpi)
        self.img_cnt = 0

        for out_port in self.g.prod_edge_map:
            out_port.producer.events['put'].append(self.intf_put)

        for in_port in self.g.cons_edge_map:
            in_port.consumer.events['ack'].append(self.intf_ack)
            in_port.consumer.events['finish'].append(self.intf_finish)

    @property
    def cur_img_file_name(self):
        return os.path.join(self.outdir, f'img_{self.img_cnt:04}.gif')

    def mark_done(self, sim):
        for sim_module in sim.sim_gears:
            if sim_module.done:
                self.g.node_map[sim_module.gear].set_style('filled')
                self.g.node_map[sim_module.gear].set_fillcolor('red')

    def color_node(self, node, state=None):
        if state:
            node.set_style('filled')
            node.set_fillcolor(self.node_state_color_map[state])
        elif self.get_node_state(node) != 'finished':
            node.set_style('')

    def get_node_state(self, node):
        return list(self.node_state_color_map.keys())[list(
            self.node_state_color_map.values()).index(node.get_fillcolor())]

    def color_edge(self, edge, state=None):
        if state:
            edge.set_color(self.node_state_color_map[state])
            edge.set_penwidth(6)
        else:
            if self.get_edge_state(edge) != 'finished':
                edge.set_color('black')
                edge.set_penwidth(1)

            edge.set_taillabel('')

    def get_edge_state(self, edge):
        return list(self.edge_state_color_map.keys())[list(
            self.edge_state_color_map.values()).index(edge.get_color())]

    def intf_finish(self, intf):
        if intf.producer:
            self.color_edge(self.g.cons_edge_map[intf.producer], 'finished')

    def intf_put(self, intf, val):
        for port in intf.end_consumers:
            edge = self.g.cons_edge_map[port]
            self.color_edge(edge, 'forward')
            edge.set_taillabel(f'< <B>{str(val)}</B> >')

    def intf_ack(self, intf):
        self.color_edge(self.g.cons_edge_map[intf.producer], 'back')

    def after_call_forward(self, sim, sim_gear):
        module = sim_gear.gear
        self.color_node(self.g.node_map[module], 'forward')

        self.mark_done(sim)

        self.snapshot()

        self.color_node(self.g.node_map[module], None)

        return True

    def after_call_back(self, sim, sim_gear):
        module = sim_gear.gear
        self.color_node(self.g.node_map[module], 'back')

        self.mark_done(sim)

        for port in module.out_ports:
            edges = self.g.prod_edge_map[port]
            for e in edges:
                self.color_edge(e, None)

        self.snapshot()

        self.color_node(self.g.node_map[module], None)

        return True

    def snapshot(self):
        self.g.write_gif(self.cur_img_file_name)
        self.img_cnt += 1

    def after_cleanup(self, sim):
        self.mark_done(sim)
        self.snapshot()



import ast
from pygears.typing import Any, Fixpnumber, Tuple, Uint, Unit
from . import ir, Context
from .call import parse_func_call
from pygears.typing import get_match_conds, TypeMatchError


def tuple_mul_resolver(opexp, ctx: Context):
    if not isinstance(opexp[1], ir.ResExpr):
        raise TypeMatchError

    return ir.ConcatExpr(opexp[0].operands * int(opexp[1].val))


def concat_resolver(opexp, ctx: Context):
    ops = tuple(op for op in reversed(opexp) if op.dtype.width)

    if len(ops) == 0:
        return ir.ResExpr(Unit())
    elif len(ops) == 1:
        return ops[0]
    else:
        tuple_res = ir.ConcatExpr(ops)
        return ir.CastExpr(tuple_res, Uint[tuple_res.dtype.width])


def fixp_add_resolver(opexp, ctx: Context):
    t_op1 = opexp[0].dtype
    t_op2 = opexp[1].dtype

    t_sum = t_op1 + t_op2
    t_cast = t_sum.base[t_sum.integer - 1, t_sum.width - 1]
    sh1 = t_sum.fract - t_op1.fract
    sh2 = t_sum.fract - t_op2.fract

    if sh1 or sh2:

        def fixp__add__(op1: t_op1, op2: t_op2) -> t_sum:
            return t_cast(op1) + t_cast(op2)

        return parse_func_call(fixp__add__, opexp, {}, ctx)
    else:
        return ir.BinOpExpr(opexp, ir.opc.Add)


def fixp_sub_resolver(opexp, ctx: Context):
    t_op1 = opexp[0].dtype
    t_op2 = opexp[1].dtype

    t_sum = t_op1 - t_op2
    t_cast = t_sum.base[t_sum.integer - 1, t_sum.width - 1]
    sh1 = t_sum.fract - t_op1.fract
    sh2 = t_sum.fract - t_op2.fract

    if sh1 or sh2:

        def fixp__sub__(op1: t_op1, op2: t_op2) -> t_sum:
            return t_cast(op1) - t_cast(op2)

        return fixp__sub__
    else:
        return ir.BinOpExpr(opexp, ir.opc.Sub)


resolvers = {
    ast.MatMult: {
        Any: concat_resolver
    },
    ast.Mult: {
        Tuple: tuple_mul_resolver
    },
    ast.Add: {
        Fixpnumber: fixp_add_resolver
    },
    ast.Sub: {
        Fixpnumber: fixp_sub_resolver
    }
}


def resolve_arith_func(op, opexp, ctx: Context):
    if type(op) in resolvers:
        op_resolvers = resolvers[type(op)]
        for templ in op_resolvers:
            try:
                try:
                    get_match_conds(opexp[0].dtype, templ)
                except AttributeError:
                    breakpoint()

                return op_resolvers[templ](opexp, ctx)
            except TypeMatchError:
                continue

    finexpr = ir.BinOpExpr((opexp[0], opexp[1]), type(op))
    for opi in opexp[2:]:
        finexpr = ir.BinOpExpr((finexpr, opi), type(op))

    return finexpr



import typing
from pygears.typing import Queue, Union, typeof, Tuple
from pygears.typing.math import ceil_chunk, ceil_pow2, bitw, ceil_div
from dataclasses import dataclass, field


def index_to_sv_slice(dtype, key):
    subtype = dtype[key]

    if isinstance(key, str):
        key = dtype.fields.index(key)

    if isinstance(key, slice):
        key = min(key.start, key.stop)

    if key is None or key == 0:
        low_pos = 0
    else:
        low_pos = dtype[:key].width

    high_pos = low_pos + subtype.width - 1

    return f'{high_pos}:{low_pos}'


@dataclass
class AxiPortConf:
    parent: str
    t: str
    port: typing.Any = None
    datamap: typing.Any = None
    params: dict = field(default_factory=dict)

    def __post_init__(self):
        if isinstance(self.parent, AxiIntfConf):
            id_param_name = f'C_AXI_{self.parent.name.upper()}_ID_WIDTH - 1'
            if self.t == 'araddr':
                self.params['arid'] = id_param_name
            elif self.t == 'awaddr':
                self.params['awid'] = id_param_name
            elif self.t == 'rdata':
                self.params['rid'] = id_param_name
            elif self.t == 'bresp':
                self.params['bid'] = id_param_name

    def dataslice(self, signal):
        if not self.datamap:
            return f'{self.port.dtype.width-1}:0'

        if not isinstance(self.datamap, dict):
            return index_to_sv_slice(self.port.dtype, self.datamap)

        return index_to_sv_slice(self.port.dtype, self.datamap[signal])

    @property
    def name(self):
        return self.port.basename


@dataclass
class AxiIntfConf:
    name: str
    t: str
    comp: typing.Dict[str, AxiPortConf] = field(default_factory=dict)

    @property
    def direction(self):
        d = ''

        if 'wdata' in self.comp:
            d += 'w'

        if 'rdata' in self.comp:
            d += 'r'

        if self.t == 'axis' and 'tdata' in self.comp:
            if self.comp['tdata'].port.direction == 'in':
                d += 'w'
            else:
                d += 'r'

        return d


def apply_key(key, dtype):
    if isinstance(key, tuple):
        for k in key:
            dtype = dtype[k]

    else:
        dtype = dtype[key]

    return dtype


def port_conf(parent, type_, p, datamap=None):
    if datamap is None:
        if not typeof(p.dtype, Queue) or type_ not in ['wdata', 'rdata', 'tdata']:
            width = p.dtype.width
        else:
            width = p.dtype['data'].width
            if type_ == 'rdata':
                datamap = {'rdata': 'data', 'rlast': 'eot'}
            elif type_ == 'wdata':
                datamap = {'wdata': 'data', 'wlast': 'eot'}
            elif type_ == 'tdata':
                datamap = {'tdata': 'data', 'tlast': 'eot'}

    else:
        key = datamap
        if isinstance(key, dict):
            key = key[type_]

        dtype = apply_key(key, p.dtype)

        width = dtype.width

    if type_ in ['wdata', 'rdata']:
        width = ceil_chunk(ceil_pow2(width), 32)
    elif type_ == 'tdata':
        width = ceil_chunk(ceil_pow2(width), 8)

    conf = AxiPortConf(parent, type_, p, datamap)

    if type_ == 'wdata':
        conf.params['wdata'] = width
        conf.params['wstrb'] = width // 8
    else:
        conf.params[type_] = width

    if type_ == 'tdata' and datamap and 'tlast' in datamap:
        conf.params['tlast'] = p.dtype.eot.width

    return conf


def get_port_def(top, name, axi_name, subintf, parent, axi_conf):
    for p in top.in_ports + top.out_ports:
        if p.basename == name:
            break
    else:
        breakpoint()
        raise Exception(
            f'Port "{name}" supplied for {subintf} port of the'
            f' {axi_name} interface, not found')

    if axi_conf['type'] == 'axi':
        if p.direction == 'in' and subintf not in ['araddr', 'awaddr', 'wdata']:
            raise Exception(
                f'Cannot drive gear port {name} from AXi port {axi_name}.{subintf}')

        if p.direction == 'out' and subintf not in ['rdata']:
            raise Exception(
                f'Cannot drive AXi port {axi_name}.{subintf} from gear port {name}')

    if subintf == 'awaddr':
        if typeof(p.dtype, Tuple) and axi_conf.get('wdata', '') == name:
            return port_conf(parent, subintf, p, 0)

    if subintf == 'araddr':
        if typeof(p.dtype, Tuple) and axi_conf.get('wdata', '') == name:
            return port_conf(parent, subintf, p, 0)

    if subintf == 'wdata':
        if typeof(p.dtype, Tuple) and axi_conf.get('awaddr', '') == name:
            if typeof(p.dtype[1], Union) and axi_conf.get('araddr', '') == name:
                return port_conf(parent, subintf, p, (1, 0))
            else:
                return port_conf(parent, subintf, p, 1)

    return port_conf(parent, subintf, p)


def port_by_name(top, name):
    for p in top.in_ports + top.out_ports:
        if p.basename == name:
            return p

    return None


def get_axi_conf(top, conf):
    axi_port_cfg = {}

    for name, pconf in conf.items():
        if isinstance(pconf, str):
            p = port_by_name(top, name)
            if p is None:
                raise Exception(f'Port "{name}" not found in the module')

            if p.direction == 'in':
                if pconf == 'axi':
                    pconf = {'type': pconf, 'wdata': name}
                elif pconf == 'axidma':
                    pconf = {'type': pconf, 'rdata': name}
                elif pconf == 'axis':
                    pconf = {'type': pconf, 'tdata': name}

            elif p.direction == 'out':
                if pconf == 'axi':
                    pconf = {'type': pconf, 'rdata': name}
                elif pconf == 'axidma':
                    pconf = {'type': pconf, 'wdata': name}
                elif pconf == 'axis':
                    pconf = {'type': pconf, 'tdata': name}

        if pconf['type'] not in ['axi', 'axidma', 'bram', 'axis']:
            continue

        if pconf['type'] in ['axi', 'axidma', 'bram']:
            axi_port_cfg[name] = AxiIntfConf(name, pconf['type'])
            for subintf in ['araddr', 'rdata', 'awaddr', 'wdata']:
                if subintf not in pconf:
                    continue

                axi_port_cfg[name].comp[subintf] = get_port_def(
                    top, pconf[subintf], name, subintf, axi_port_cfg[name], pconf)

        elif pconf['type'] in ['axis']:
            axi_port_cfg[name] = AxiIntfConf(name, pconf['type'])
            for subintf in ['tdata']:
                if subintf not in pconf:
                    continue

                axi_port_cfg[name].comp[subintf] = get_port_def(
                    top, pconf[subintf], name, subintf, axi_port_cfg[name], pconf)


    for name, p in axi_port_cfg.copy().items():
        if p.t in ['axidma']:
            if 'rdata' in p.comp:
                p.comp['araddr'] = AxiPortConf(p, 'araddr', params={'araddr': 32})

            if 'wdata' in p.comp:
                p.comp['awaddr'] = AxiPortConf(p, 'awaddr', params={'awaddr': 32})

            axil_comp = {
                'awaddr': AxiPortConf('axilite', 'awaddr', params={'awaddr': 5}),
                'wdata':
                AxiPortConf('axilite', 'wdata', params={
                    'wdata': 32,
                    'wstrb': 4
                }),
                'araddr': AxiPortConf('axilite', 'araddr', params={'araddr': 5}),
                'rdata': AxiPortConf('axilite', 'rdata', params={'rdata': 32}),
                'bresp': AxiPortConf('axilite', 'bresp', params={'bresp': True})
            }

            axi_port_cfg[f'{name}_ctrl'] = AxiIntfConf(
                f'{name}_ctrl', 'axilite', axil_comp)

            if 'wdata' in p.comp and 'bresp' not in p.comp:
                p.comp['bresp'] = AxiPortConf(
                    p, 'bresp', params={'bresp': 'awaddr' in p.comp})

        if p.t in ['axi']:
            if 'wdata' in p.comp and 'awaddr' not in p.comp:
                p.comp['awaddr'] = AxiPortConf(p, 'awaddr', params={'awaddr': 3})

            if 'rdata' in p.comp and 'araddr' not in p.comp:
                p.comp['araddr'] = AxiPortConf(p, 'araddr', params={'araddr': 3})

        if p.t in ['axi', 'bram']:
            if 'wdata' in p.comp and 'awaddr' in p.comp:
                wdata_byte_addr_w = bitw(ceil_div(p.comp['wdata'].params['wdata'], 8) - 1)
                p.comp['awaddr'].params['awaddr'] += wdata_byte_addr_w

            if 'rdata' in p.comp and 'araddr' in p.comp:
                rdata_byte_addr_w = bitw(ceil_div(p.comp['rdata'].params['rdata'], 8) - 1)
                p.comp['araddr'].params['araddr'] += rdata_byte_addr_w

            if 'wdata' in p.comp and 'bresp' not in p.comp:
                p.comp['bresp'] = AxiPortConf(
                    p, 'bresp', params={'bresp': 'awaddr' in p.comp})

    return axi_port_cfg



import subprocess
from pygears import PluginBase, reg, find
from pygears.core.graph import get_consumer_tree
from pygears.core.port import OutPort
from pygears.sim import log, timestep, SimPlugin
from pygears.sim.sim_gear import SimGear
from pygears.typing import typeof, TLM, Float, Uint
from pygears.typing.visitor import TypingVisitorBase
from vcd import VCDWriter
from pygears.core.hier_node import HierVisitorBase
from .sim_extend import SimExtend
import os
import fnmatch
import itertools
import atexit
from pygears.conf import inject, Inject


def match(val, include_pattern):
    def check(pattern):
        if isinstance(pattern, str):
            return fnmatch.fnmatch(val.name, pattern)
        else:
            return pattern(val)

    return any(check(p) for p in include_pattern)


def is_trace_included(port, include):
    return True
    # # if not match(f'{port.gear.name}.{port.basename}', include):
    # if not match(port, include):
    #     return False

    # if (port.dtype is None) or (typeof(port.dtype, TLM) and not vcd_tlm):
    #     return False

    # return True


class VCDHierVisitor(HierVisitorBase):
    @inject
    def __init__(self, include, sim_map=Inject('sim/map')):
        self.include = include
        self.sim_map = sim_map
        self.vcd_vars = set()
        self.end_consumers = {}

    def trace_if_included(self, p):
        if not is_trace_included(p, self.include):
            return

        self.vcd_vars.add(p)

    def Gear(self, module):
        if module.parent is None:
            return super().HierNode(module)

        for p in module.in_ports:
            self.trace_if_included(p)

        if module in self.sim_map or module.hierarchical:
            for p in module.out_ports:
                self.trace_if_included(p)

        if module in self.sim_map:
            for p in module.in_ports:
                # TODO Hack to make end_consumer unique with id(intf) so that
                # it can be looked upon in the list. Make it in a better way
                self.end_consumers[p.consumer] = {'prods': [], 'intf': id(p.consumer)}

        if (module in self.sim_map or module.hierarchical) and module.params['sim_cls'] is None:
            super().HierNode(module)

        return True


from enum import IntEnum


class ChannelState(IntEnum):
    Invalid = 0
    NotReady = 1
    Ready = 2
    Done = 3


class WebSim(SimExtend):
    @inject
    def __init__(self,
                 trace_fn='pygears.json',
                 include=Inject('debug/trace'),
                 sim=Inject('sim/simulator'),
                 outdir=Inject('results-dir')):

        super().__init__()
        self.sim = sim
        self.finished = False
        self.outdir = outdir
        self.trace_fn = None
        self.shmid_proc = None
        self.include = include

        self.trace_fn = os.path.abspath(os.path.join(self.outdir, trace_fn))
        atexit.register(self.finish)

        try:
            subprocess.call(f"rm -f {self.trace_fn}", shell=True)
        except OSError:
            pass

        log.info(f'Main VCD dump to "{self.trace_fn}"')
        # self.json_vcd = open(self.trace_fn, 'w')

        self.json_vcd = {}

        self.handhake = set()
        self.valid = set()
        self.value_change = {}
        self.prev_cycle_handshake = set()

    def before_run(self, sim):
        vcd_visitor = VCDHierVisitor(self.include)
        vcd_visitor.visit(find('/'))
        self.vcd_vars = {
            p: {
                'srcs': [],
                'srcs_active': [],
                'dtype': p.dtype,
                'val': None,
            }
            for p in vcd_visitor.vcd_vars
        }

        for p in self.vcd_vars:
            self.json_vcd[p.name] = []

        if not vcd_visitor.vcd_vars:
            self.deactivate('before_run')
            return True

        self.end_consumers = vcd_visitor.end_consumers

        for intf in self.end_consumers:
            intf.events['put'].append(self.intf_put)
            intf.events['ack'].append(self.intf_ack)

        self.extend_intfs()

    def extend_intfs(self):
        for p, v in self.vcd_vars.items():
            v['srcs'] = [self.end_consumers[pp.consumer] for pp in get_consumer_tree(p.consumer)]
            v['srcs_active'] = [False] * len(v['srcs'])
            v['p'] = p
            for vs in v['srcs']:
                vs['prods'].append(v)

        reg['graph/consumer_tree'] = {}
        reg['graph/end_producer'] = {}

    def var_put(self, v, val):
        self.valid.add(v['p'])
        self.value_change[v['p']] = val

        # print(f'{v["p"].name}: {val}')
        # breakpoint()
        # if typeof(v['dtype'], TLM):
        #     self.writer.change(v['data'], timestep() * 10, str(val))
        # else:
        #     try:
        #         visitor = VCDValVisitor(v, self.writer, timestep() * 10, max_level=10)
        #         visitor.visit(v['dtype'], 'data', val=val)
        #     except AttributeError:
        #         pass

        # self.writer.change(v['valid'], timestep() * 10, 1)

    def intf_put(self, intf, val):
        p = intf.producer
        if p in self.vcd_vars:
            v = self.vcd_vars[p]
            self.var_put(v, val)

        if intf in self.end_consumers:
            v = self.end_consumers[intf]
            for vp in v['prods']:
                if not any(vp['srcs_active']):
                    # TODO: Optimization possibility, don't write the data, only ready/valid signals
                    self.var_put(vp, val)

                for i, vv in enumerate(vp['srcs']):
                    if vv is v:
                        vp['srcs_active'][i] = True
                        break

        return True

    def intf_ack(self, intf):
        p = intf.producer
        if p in self.vcd_vars:
            v = self.vcd_vars[p]
            # print(f'{v["p"].name}: HSHK')
            # self.writer.change(v['ready'], timestep() * 10, 1)
            self.handhake.add(p)

        if intf in self.end_consumers:
            v = self.end_consumers[intf]
            for vp in v['prods']:

                for i, vv in enumerate(vp['srcs']):
                    if vv is v:
                        vp['srcs_active'][i] = False
                        break

                if not any(vp['srcs_active']):
                    # print(f'{vp["p"].name}: HSHK')
                    # self.writer.change(vp['ready'], timestep() * 10, 1)
                    self.handhake.add(vp['p'])

        return True

    def create_data_change(self, val, prev_val):
        from pygears.typing import Queue, Array, Tuple
        is_changed = prev_val is None and not val is None or val != prev_val

        if isinstance(val, (Queue, Array, Tuple)):
            if prev_val is None:
                prev_val = [None] * len(val)

            change = [self.create_data_change(v, prev_v) for v, prev_v in zip(val, prev_val)]
            return {'isValueComplex': True, 'isDataChanged': is_changed, 'value': change}
        else:
            if isinstance(val, (int, float)):
                val = int(val)
            else:
                val = val.code()

            return {'isValueComplex': False, 'isDataChanged': is_changed, 'value': val}

    def create_change(self, timestep, state, state_change, val, prev_val):
        elem = {
            'cycle': timestep,
            'state': int(state),
            'isStateChanged': state_change,
        }

        if val is not None:
            elem['data'] = self.create_data_change(val, prev_val)

        return elem

    def after_timestep(self, sim, timestep):
        for p, v in self.vcd_vars.items():
            changes = self.json_vcd[p.name]
            ack = p in self.handhake and not any(v['srcs_active'])
            if ack:
                state = ChannelState.Ready
            elif p in self.valid:
                state = ChannelState.NotReady
            else:
                state = ChannelState.Invalid

            became_not_ready = p in self.prev_cycle_handshake and state != ChannelState.Ready
            became_valid = p in self.valid and v['val'] is None

            state_change = ((timestep == 0 and state != ChannelState.Invalid)
                            or (ack and p not in self.prev_cycle_handshake) or became_valid
                            or became_not_ready)
            data_change = False
            if p in self.value_change:
                data_change = v['val'] is None or (p in self.prev_cycle_handshake
                                                   and self.value_change[p] != v['val'])

            if state_change or data_change or timestep == 0:
                cycle_change = self.create_change(timestep, state, state_change,
                                                  self.value_change.get(p, v['val']), v['val'])
                if cycle_change is not None:
                    changes.append(cycle_change)

            if state == ChannelState.Ready:
                v['val'] = self.value_change[p]
                del self.value_change[p]
                self.valid.remove(p)
            elif state == ChannelState.NotReady:
                v['val'] = self.value_change[p]
            elif state == ChannelState.Invalid:
                v['val'] = None

            if p in self.handhake and any(v['srcs_active']):
                self.handhake.remove(p)

            # cycle_change = None
            # if p in self.valid:
            #     ack = p in self.handhake and not any(v['srcs_active'])
            #     state = ChannelState.Ready if ack else ChannelState.NotReady
            #     state_change = ack and p not in self.prev_cycle_handshake
            #     data_change = v['val'] is None or self.value_change[p] != v['val']
            #     if state_change or data_change:
            #         cycle_change = self.create_change(timestep, state, state_change,
            #                                           self.value_change[p], v['val'])
            # elif p in self.handhake:
            #     cycle_change = self.create_change(timestep, ChannelState.Ready,
            #                                       self.value_change[p], v['val'])

            # elif p in self.prev_cycle_handhake:
            #     cycle_change = self.create_change(timestep, ChannelState.Invalid, None, None)
            # else:
            #     v['val'] = None

        self.prev_cycle_handshake = self.handhake.copy()
        self.handhake = set()

        return True

    def finish(self):
        if not self.finished:
            json_out = {'startCycle': 0, 'endCycle': timestep(), 'channelChanges': []}

            for p in self.vcd_vars:
                json_out['channelChanges'].append({
                    'channelName': p.producer.name,
                    'changes': self.json_vcd[p.name]
                })

            import json
            # json.dump(json_out, open(self.trace_fn, 'w'), indent=4)
            json.dump(json_out, open(self.trace_fn, 'w'))
            self.finished = True

    def after_cleanup(self, sim):
        self.finish()



import fnmatch
import functools
from collections import OrderedDict

from pygears import reg
from pygears.core.gear import OutSig
# from .inst import SVGenInstPlugin
# from pygears.hdl.sv.svparse import parse
from pygears.hdl.modinst import HDLModuleInst
from pygears.hdl import hdl_log

from .vcompile import compile_gear
from ..sv.svparse import parse
from ..sv.sv_keywords import sv_keywords as keywords


class VModuleInst(HDLModuleInst):
    def __init__(self, node):
        super().__init__(node, 'v')
        self.vgen_map = reg['vgen/map']

    @property
    def inst_name(self):
        inst_name = super().inst_name

        if inst_name in keywords:
            return f'{inst_name}_i'
        else:
            return inst_name

    @property
    @functools.lru_cache()
    def traced(self):
        self_traced = any(
            fnmatch.fnmatch(self.node.name, p)
            for p in reg['debug/trace'])

        if self.is_hierarchical:
            children_traced = any(self.vgen_map[child].traced
                                  for child in self.node.child)
        else:
            children_traced = False

        return self_traced or children_traced

    @functools.lru_cache()
    def impl_parse(self):
        if self.impl_path:
            # raise Exception('parse not implemented')
            with open(self.impl_path, 'r') as f:
                return parse(f.read())
        else:
            hdl_log().warning(f'Verilog file not found for {self.node.name}')

    def get_synth_wrap(self, template_env):
        context = {
            'module_name': self.module_name,
            'inst_name': self.inst_name,
            'intfs': list(self.port_configs),
            'sigs': self.node.meta_kwds['signals'],
            'param_map': self.params
        }
        return template_env.render_local(__file__, "module_synth_wrap.j2",
                                         context)

    def get_out_port_map_intf_name(self, port):
        return self.vgen_map[port.consumer].basename, None, None

    def get_in_port_map_intf_name(self, port):
        intf = port.producer
        vgen_intf = self.vgen_map[intf]

        if len(intf.consumers) == 1:
            return vgen_intf.outname, None, None
        else:
            return (vgen_intf.outname, intf.consumers.index(port),
                    intf.dtype.width)

    def get_compiled_module(self, template_env):
        return compile_gear(self.node.gear, template_env, self.module_context)

    def get_hier_module(self, template_env):
        context = self.module_context

        self.vgen_map = reg['vgen/map']

        for child in self.node.local_interfaces():
            vgen = self.vgen_map[child]
            contents = vgen.get_inst(template_env)
            if contents:
                context['inst'].append(contents)

        for child in self.node.local_modules():
            for s in child.meta_kwds['signals']:
                if isinstance(s, OutSig):
                    name = child.params['sigmap'][s.name]
                    context['inst'].append(f'logic [{s.width-1}:0] {name};')

            vgen = self.vgen_map[child]
            if hasattr(vgen, 'get_inst'):
                contents = vgen.get_inst(template_env)
                if contents:
                    if vgen.traced:
                        context['inst'].append('/*verilator tracing_on*/')
                    context['inst'].append(contents)
                    if vgen.traced:
                        context['inst'].append('/*verilator tracing_off*/')

        return template_env.render_local(__file__, "hier_module.j2", context)

    def get_inst(self, template_env):
        param_map = self.params

        port_map = {
            port.basename: self.get_in_port_map_intf_name(port)
            for port in self.node.in_ports
        }

        port_map.update({
            port.basename: self.get_out_port_map_intf_name(port)
            for port in self.node.out_ports
        })

        context = {
            'rst_name': 'rst',
            'module_name': self.module_name,
            'inst_name': self.inst_name,
            'param_map': param_map,
            'port_map': port_map,
            'sig_map': self.node.params['sigmap']
        }

        return template_env.snippets.module_inst(**context)




from pygears import gear, alternative
from pygears.typing import Queue, Uint


@gear(enablement=b'running==False')
async def qcnt(din: Queue, *, running=False, lvl=0, init=1,
               w_out=16) -> Uint['w_out']:
    """Short for Queue Length Count. Counts the number of input sub-transactions
    and outputs the final count when the input transaction finishes.

    Args:
        lvl: Specifies the level of the input Queue that needs to be counted
        init: Initialization value for the counter
        w_out: Width of the counter and the output data bus

    Returns:
        The number of sub-transactions in the input transaction.
    """

    cnt = Uint[w_out](init)

    async for (data, eot) in din:
        if all(eot):
            yield cnt

        if all(eot[:lvl]):
            cnt += 1


@alternative(qcnt)
@gear(enablement=b'running==True')
async def qcnt_running(din: Queue, *, running=True, lvl=0, init=1,
                       w_out=16) -> Queue[Uint['w_out']]:
    """Short for Queue Count. Counts the number of sub-transactions in the input
    transaction and outputs the running count.

    Args:
        lvl: Specifies the level of the input Queue that needs to be counted
        init: Initialization value for the counter
        w_out: Width of the counter and the output data bus

    Returns:
        The running count of sub-transactions in the input transaction.
    """
    cnt = Uint[w_out](init)

    async for (data, eot) in din:
        if all(eot[:lvl]):
            yield (cnt, all(eot))
            cnt += 1



from pygears.conf import PluginBase, reg
from . import ir
from .ast.cast import resolve_cast_func
from .ast.call import resolve_gear_call, resolve_func
from pygears import Intf, reg

from pygears.core.gear import OutSig

from functools import reduce
from pygears.typing import Int, Uint, code, div, Queue, Integral, Float, Union, Bool, Maybe
from pygears.typing import is_type, typeof, Tuple, Array
from pygears.typing import floor, cast, signed, saturate
from pygears.typing.queue import QueueMeta

from pygears.util.utils import gather, qrange
from pygears.sim import clk
from pygears.lib.rng import qrange as qrange_gear
from pygears.lib.saturate import saturate as saturate_gear


def call_div(a, b, subprec):
    t_a = a.dtype
    t_b = b.dtype

    t_div = div(t_a, t_b, int(subprec.val))

    def fixp__div__(op1: t_a, op2: t_b) -> t_div:
        return t_div(op1) // op2

    return fixp__div__


def max_expr(op1, op2):
    op1_compare = op1
    op2_compare = op2

    # TODO: Sort this casting out
    signed = typeof(op1.dtype, Int) or typeof(op2.dtype, Int)
    if signed and typeof(op1.dtype, Uint):
        op1_compare = resolve_cast_func(op1, Int)
    if signed and typeof(op2.dtype, Uint):
        op2_compare = resolve_cast_func(op2, Int)

    cond = ir.BinOpExpr((op1_compare, op2_compare), ir.opc.Gt)
    return ir.ConditionalExpr(cond=cond, operands=(op1, op2))


# TODO: Why do we need ir.TupleExpr?
def call_tuple(arg):
    if isinstance(arg, ir.ConcatExpr):
        return ir.TupleExpr(arg.operands)
    elif isinstance(arg, ir.TupleExpr):
        return arg
    elif isinstance(arg, ir.ResExpr):
        # TODO: Array is a list, so we have a workaround here
        if typeof(arg.dtype, (Tuple, Array, Queue)):
            return arg
        else:
            return ir.ResExpr(tuple(arg.val))
    elif typeof(arg.dtype, (Array, Tuple)):
        return ir.ConcatExpr([ir.SubscriptExpr(arg, ir.ResExpr(i)) for i in range(len(arg.dtype))])
    else:
        breakpoint()
        raise Exception

def call_tuple_add(op1, op2):
    if op1 == ir.ResExpr(()):
        return op2
    elif op2 == ir.ResExpr(()):
        return op1

    if not isinstance(op1, ir.ConcatExpr):
        op1 = call_tuple(op1)
    if not isinstance(op2, ir.ConcatExpr):
        op2 = call_tuple(op2)

    if isinstance(op1, ir.ConcatExpr):
        ops1 = op1.operands
    elif isinstance(op1, ir.ResExpr):
        ops1 = [ir.ResExpr(v) for v in op1.val]

    if isinstance(op2, ir.ConcatExpr):
        ops2 = op2.operands
    elif isinstance(op2, ir.ResExpr):
        ops2 = [ir.ResExpr(v) for v in op2.val]

    return ir.ConcatExpr(ops1 + ops2)


def call_len(arg, **kwds):
    if isinstance(arg, ir.ConcatExpr):
        return ir.ResExpr(len(arg.operands))

    if isinstance(arg, ir.ResExpr):
        return ir.ResExpr(len(arg.val))

    return ir.ResExpr(len(arg.dtype))


def call_print(*arg, **kwds):
    pass


def call_float(arg, **kwds):
    return ir.CastExpr(arg, cast_to=Float)


def call_int(arg, **kwds):
    # ignore cast
    if typeof(arg.dtype, (Uint, Int)):
        return arg
    elif typeof(arg.dtype, Integral):
        if arg.dtype.signed:
            return ir.CastExpr(arg, cast_to=Int[arg.dtype.width])
        else:
            return ir.CastExpr(arg, cast_to=Uint[arg.dtype.width])
    else:
        return ir.ResExpr(NotImplemented)


def call_all(arg, **kwds):
    return ir.ArrayOpExpr(arg, ir.opc.BitAnd)


def call_any(arg, **kwds):
    return ir.ArrayOpExpr(arg, ir.opc.BitOr)


# TODO: Can this be generalized a bit?
def call_max(*arg, **kwds):
    if len(arg) != 1:
        return reduce(max_expr, arg)

    arg = arg[0]

    assert typeof(arg.dtype, Tuple), 'Not supported yet...'

    op = []
    for field in arg.dtype.fields:
        op.append(ir.SubscriptExpr(arg, ir.ResExpr(field)))

    return reduce(max_expr, op)


def call_uint_matmul(obj, arg):
    return ir.CastExpr(ir.ConcatExpr(operands=[arg, obj]), cast_to=(obj.dtype @ arg.dtype))


def call_sub(obj, arg):
    return ir.CastExpr(arg, cast_to=obj.sub())


def outsig_write(obj, arg):
    return ir.AssignValue(obj, arg)


def call_get(obj, *args, **kwds):
    return obj


def call_get_nb(obj, *args, **kwds):
    return ir.AssignValue(ir.Component(obj, 'ready'), ir.res_true)
    # return obj


def call_pull_nb(obj):
    return ir.Component(obj, 'data')


def call_put_nb(obj, arg):
    ctx = reg['hls/ctx'][-1]

    if isinstance(obj, Intf):
        for p in ctx.gear.out_ports:
            if p.producer is obj:
                obj = ctx.ref(p.basename)
                break
        else:
            breakpoint()

    return [
        ir.Await('forward'),
        ir.AssignValue(ir.Component(obj, 'data'), arg),
        ir.AssignValue(ir.Component(obj, 'valid'), ir.res_true)
    ]
    # return obj


def call_clk(*arg, **kwds):
    return None


def call_empty(obj):
    return ir.UnaryOpExpr(ir.Component(obj, 'valid'), ir.opc.Not)


def call_ack(obj):
    return ir.AssignValue(ir.Component(obj, 'ready'), ir.res_true)


def call_gather(*arg, **kwds):
    return ir.ConcatExpr(operands=list(arg))


def call_cast(arg, cast_type):
    return resolve_cast_func(arg, cast_type.val)


def call_signed(val):
    if val.dtype.signed:
        return val

    if typeof(val.dtype, Uint):
        return resolve_cast_func(val, Int)

    raise Exception("Unsupported signed cast")


def call_code(val, cast_type=ir.ResExpr(Uint)):
    cast_type = code(val.dtype, cast_type.val)
    if val.dtype == cast_type:
        return val

    return ir.CastExpr(val, cast_to=cast_type)


def call_type(arg):
    return ir.ResExpr(arg.dtype)


def call_isinstance(arg, dtype):
    if isinstance(dtype, ir.ResExpr):
        dtype = dtype.val

    if isinstance(arg, ir.ResExpr):
        return Bool(isinstance(arg.val, dtype))

    return ir.ResExpr(Bool(typeof(arg.dtype, dtype)))


def call_is_type(arg):
    if not isinstance(arg, ir.ResExpr):
        return ir.res_false

    return ir.ResExpr(Bool(is_type(arg.val)))


def call_typeof(arg, dtype):
    if isinstance(dtype, ir.ResExpr):
        dtype = dtype.val

    if not isinstance(arg, ir.ResExpr):
        return ir.res_false

    return ir.ResExpr(Bool(typeof(arg.val, dtype)))


# TODO: Implement sum and reduce
def call_sum(iterable, start):
    if not isinstance(iterable, ir.ConcatExpr):
        breakpoint()

    f = getattr(start.dtype, '__add__')
    ret = resolve_func(f, (start, iterable.operands[0]), {}, reg['hls/ctx'])

    if len(iterable.operands) == 1:
        return ret
    else:
        return call_sum(ir.ConcatExpr(iterable.operands[1:]), ret)


def call_subs_fix_index(orig, path, val):
    parts = []
    for i in range(len(orig.dtype)):
        p = ir.SubscriptExpr(orig, ir.ResExpr(i))
        if isinstance(path, tuple):
            if path[0].val == i:
                p = ir.CastExpr(call_subs(p, path[1:], val), p.dtype)
        elif path.val == i:
            p = ir.CastExpr(val, p.dtype)

        parts.append(p)

    return ir.CastExpr(ir.ConcatExpr(parts), orig.dtype)


def call_subs_var_index(orig, path, val):
    parts = []
    for i in range(len(orig.dtype)):
        p = ir.SubscriptExpr(orig, ir.ResExpr(i))
        if isinstance(path, tuple):
            cond = ir.BinOpExpr((path[0], ir.ResExpr(i)), ir.opc.Eq)
            repl = ir.CastExpr(call_subs(p, path[1:], val), p.dtype)
        else:
            cond = ir.BinOpExpr((path, ir.ResExpr(i)), ir.opc.Eq)
            repl = ir.CastExpr(val, p.dtype)

        parts.append(ir.ConditionalExpr((repl, p), cond))

    return ir.CastExpr(ir.ConcatExpr(parts), orig.dtype)


def call_subs(orig, *args, **kwds):
    if args:
        path, val = args
        if isinstance(path, tuple) and len(path) == 1:
            path = path[0]

        if isinstance(path, tuple) and isinstance(path[0], ir.ResExpr):
            return call_subs_fix_index(orig, path, val)
        elif not isinstance(path, tuple) and isinstance(path, ir.ResExpr):
            return call_subs_fix_index(orig, path, val)
        else:
            return call_subs_var_index(orig, path, val)

    if kwds:
        parts = []
        for i, name in enumerate(orig.dtype.fields):
            p = ir.SubscriptExpr(orig, ir.ResExpr(i))
            if name in kwds:
                p = ir.CastExpr(kwds[name], p.dtype)

            parts.append(p)

        return ir.CastExpr(ir.ConcatExpr(parts), orig.dtype)

def call_maybe_get(arg):
    return ir.CastExpr(ir.SubscriptExpr(arg, ir.ResExpr(0)), cast_to=arg.dtype.dtype)


def call_maybe_some(arg, val):
    if isinstance(arg, ir.ResExpr):
        breakpoint()
        arg = arg.dtype

    return ir.CastExpr(ir.ConcatExpr([val, ir.ResExpr(1)]), cast_to=arg)


def call_enumerate(arg):
    arg.enumerated = True
    return arg


def call_qrange(*args):
    return resolve_gear_call(qrange_gear.func, args, {})


def call_range(*args):
    ret = ir.CallExpr(range,
                      dict(zip(['start', 'stop', 'step'], args)),
                      params={'return': Queue[args[0].dtype]})

    ret.pass_eot = False
    return ret


def call_breakpoint():
    return None


def ir_builtin(func):
    reg['hls/ir_builtins'].get(func, None)


class AddIntfOperPlugin(PluginBase):
    @classmethod
    def bind(cls):
        ir_builtins = {
            gather: call_gather,
            all: call_all,
            any: call_any,
            max: call_max,
            clk: call_clk,
            float: call_float,
            int: call_int,
            len: call_len,
            print: call_print,
            type: call_type,
            isinstance: call_isinstance,
            tuple: call_tuple,
            tuple.__add__: call_tuple_add,
            sum: call_sum,
            is_type: call_is_type,
            typeof: call_typeof,
            div: call_div,
            ir.IntfType.empty: call_empty,
            ir.IntfType.ack: call_ack,
            Intf.get: call_get,
            Intf.get_nb: call_get_nb,
            Intf.put_nb: call_put_nb,
            ir.IntfType.pull_nb: call_pull_nb,
            cast: call_cast,
            signed: call_signed,
            QueueMeta.sub: call_sub,
            Maybe.get: call_maybe_get,
            Maybe.some: call_maybe_some,
            Maybe.some.__func__: call_maybe_some, # TODO: Why do we need both this and above
            object.__getattribute__(Array, 'subs').func: call_subs,
            object.__getattribute__(Tuple, 'subs').func: call_subs,
            object.__getattribute__(Uint, '__matmul__'): call_uint_matmul,
            OutSig.write: outsig_write,
            Array.code: call_code,
            Tuple.code: call_code,
            code: call_code,
            qrange: call_qrange,
            range: call_range,
            enumerate: call_enumerate
        }

        import sys
        if sys.version_info[1] >= 7:
            ir_builtins[breakpoint] = call_breakpoint

        int_unops = {
            ir.opc.Invert: '__invert__',
            ir.opc.UAdd: '__pos__',
            ir.opc.USub: '__neg__',
        }

        # TODO: User @wraps for better error reporting
        for op, name in int_unops.items():
            ir_builtins[getattr(int, name)] = lambda a, *, x=op: ir.UnaryOpExpr(call_int(a), x)

        int_binops = {
            ir.opc.Add: '__add__',
            ir.opc.BitAnd: '__and__',
            ir.opc.BitOr: '__or__',
            ir.opc.BitXor: '__xor__',
            ir.opc.Div: '__truediv__',
            ir.opc.Eq: '__eq__',
            ir.opc.Gt: '__gt__',
            ir.opc.GtE: '__ge__',
            ir.opc.FloorDiv: '__floordiv__',
            ir.opc.Lt: '__lt__',
            ir.opc.LtE: '__le__',
            ir.opc.LShift: '__lshift__',
            ir.opc.Mod: '__mod__',
            ir.opc.Mult: '__mul__',
            ir.opc.NotEq: '__ne__',
            ir.opc.RShift: '__rshift__',
            ir.opc.Sub: '__sub__'
        }

        # TODO: User @wraps for better error reporting
        for op, name in int_binops.items():
            ir_builtins[getattr(int, name)] = lambda a, b, *, x=op: ir.BinOpExpr(
                (call_int(a), b), x)

        int_binrops = {
            ir.opc.Add: '__radd__',
            ir.opc.BitAnd: '__rand__',
            ir.opc.BitOr: '__ror__',
            ir.opc.BitXor: '__rxor__',
            ir.opc.Div: '__rtruediv__',
            ir.opc.FloorDiv: '__rfloordiv__',
            ir.opc.LShift: '__rlshift__',
            ir.opc.Mod: '__rmod__',
            ir.opc.Mult: '__rmul__',
            ir.opc.RShift: '__rshift__',
            ir.opc.Sub: '__rsub__'
        }

        # TODO: User @wraps for better error reporting
        for op, name in int_binrops.items():
            # TODO: Test NotImplemented part
            def intop(a, b, *, x=op):
                a_conv = call_int(a)
                if a_conv == ir.ResExpr(NotImplemented):
                    return ir.ResExpr(NotImplemented)

                return ir.BinOpExpr((a_conv, b), x)

            ir_builtins[getattr(int, name)] = intop

        reg['hls/ir_builtins'] = ir_builtins



from pygears import gear
from pygears.lib.rom import rom

from pygears.typing import Fixp, Fixpnumber, Ufixp, Uint, bitw


@gear
def funclut(x: Fixpnumber, *, f, precision=b'x.width', dtype=None):
    '''Implement arbitrary 1 input parameter function as Lookup-table for
    integers. f is arbitrary function e.g. math.sqrt, precision is a number of
    bits the function result will be represented with,

    sqrt_lut: x | funclut(f=math.sqrt, precision=4)

    '''

    din_t = x.dtype

    step = 2**(-din_t.fract)
    w_din = len(din_t)

    def gen_vals_signed():
        for i in range(2**w_din):
            if i < (1 << (w_din - 1)):
                yield f(step * i)
            else:
                yield f(step * (i - (1 << w_din)))

    def gen_vals_unsigned():
        for i in range(2**w_din):
            yield f(step * i)

    def gen_vals():
        if din_t.signed:
            yield from gen_vals_signed()
        else:
            yield from gen_vals_unsigned()

    if dtype is None:
        float_res_list = list(gen_vals())
        vmin = min(map(round, float_res_list))
        vmax = max(map(round, float_res_list))

        if vmin < 0:
            dtype = Fixp[bitw(
                max(2 * abs(vmin), 2 * (vmax) +
                    (1 if vmax > 0 else 0))), precision]
        else:
            dtype = Ufixp[bitw(max(vmax, vmin)), precision]

        lut_list = [dtype(v) for v in float_res_list]
    else:
        lut_list = [dtype(v) for v in gen_vals()]

    dout = x >> Uint[w_din] | rom(data=lut_list, dtype=dtype)
    return dout



from pygears import alternative, gear
from pygears.typing import Queue, Tuple, typeof, Unit
from pygears.lib.shred import shred
from pygears.lib.ccat import ccat
from pygears.util.utils import quiter_async
from pygears import module
from pygears.util.utils import gather


def lvl_if_queue(t):
    if not issubclass(t, Queue):
        return 0
    else:
        return t.lvl


def cart_type(dtypes, order=None):
    if order is not None:
        dtypes = [dtypes[o] for o in order]

    arg_queue_lvl = [lvl_if_queue(d) for d in dtypes]

    base_type = Tuple[tuple(d if lvl == 0 else d[0]
                            for d, lvl in zip(dtypes, arg_queue_lvl))]

    # If there are no Queues, i.e. sum(arg_queue_lvl) == 0, the type below
    # will resolve to just base_type
    return Queue[base_type, sum(arg_queue_lvl)]


@gear(enablement=b'len(din) == 2')
async def cart_cat(*din, order=None) -> b'cart_type(din, order)':
    if order is None:
        order = range(len(din))

    async with gather(*din) as data:
        dout_data = []
        dout_eot = Unit()
        for o in order:
            d = data[o]
        # for d in data:
            if isinstance(d, Queue):
                dout_data.append(d.data)
                dout_eot = dout_eot @ d.eot
            else:
                dout_data.append(d)

        yield (dout_data, dout_eot)


@gear(enablement=b'len(din) == 2')
def cart(*din, order=None) -> b'cart_type(din, order)':
    return din | cart_sync(outsync=False) | cart_cat(order=order)


@alternative(cart)
@gear(enablement=b'len(din) > 2')
def cart_vararg(*din):
    ret = cart(din[0], din[1])
    for d in din[2:]:
        ret = cart(ret, d)

    return ret >> cart_type([d.dtype for d in din])


# TODO: Lowest eot for each uncart output needs to be shortened to 1 data using
# flattening
@gear
def uncart(din, *, dtypes):
    zdata = din[0]
    zlast = din[1:]

    def split():
        for i, d in enumerate(dtypes):
            data = zdata[i]
            if issubclass(d, Queue):
                yield ccat(data, zlast[:d.lvl]) | Queue[data.dtype, d.lvl]
            else:
                yield data

    return tuple(split())


@gear(enablement=b'len(din) == 2')
async def cart_sync(*din, outsync=True) -> b'din':
    async with din[0] as d0:
        if typeof(din[1].dtype, Queue):
            async for d1 in quiter_async(din[1]):
                yield d0, d1
        else:
            async with din[1] as d1:
                yield d0, d1


@alternative(cart_sync)
@gear
def cart_sync_vararg(*din):
    return din | cart | uncart(dtypes=[d.dtype for d in din])


@gear
def cart_sync_with(sync_in, din, *, balance=None):
    if balance:
        sync_in = sync_in | balance

    din_sync, sync_in_sync = cart_sync(din, sync_in)
    sync_in_sync | shred

    return din_sync

@gear
def cart_wrap_with(sync, din):
    din_cart = cart(sync, din)

    return ccat(din_cart['data'][1], din_cart['eot']) | Queue



import os
import re

from vcd.gtkw import GTKWSave

from pygears import find, reg
from pygears.core.port import OutPort as GearOutPort
from pygears.rtl.port import InPort, OutPort
from pygears.sim import log
from pygears.sim.extens.graphviz import graph
from pygears.sim.extens.vcd import module_sav
from pygears.sim.modules.sim_socket import SimSocket
from .activity import ActivityChecker


def _get_end_consumer_rec(intf, consumers):
    for port in intf.consumers:
        cons_intf = port.consumer

        if isinstance(port, InPort) and (not cons_intf.consumers):
            consumers.append(port)
        else:
            _get_end_consumer_rec(cons_intf, consumers)


def get_end_consumer(intf):
    consumers = []
    _get_end_consumer_rec(intf, consumers)
    return consumers


def _get_producer_rec(intf, producers):
    if isinstance(intf.producer, OutPort) or isinstance(
            intf.producer, GearOutPort):
        producers.append(intf.producer)
    else:
        _get_producer_rec(intf.producer, producers)


def get_producer(intf):
    producers = []
    _get_producer_rec(intf, producers)
    return producers


def find_target_prod(intf):
    # Who is producer gear?
    # prod_rtl_port = intf.producer
    end_p = get_producer(intf)
    if len(end_p) != 1:
        return None
    if isinstance(end_p[0], OutPort):
        prod_rtl_port = end_p[0]
        prod_rtl_node = prod_rtl_port.node
        prod_gear = prod_rtl_node.gear
    else:
        prod_gear = end_p[0].gear
    if len(prod_gear.child):
        if len(prod_gear.child) > 1:
            log.warning(
                f'ActivityCosim: prod has more than one child. Setting on first.'
            )
        return prod_gear.child[0]
    else:
        return prod_gear


def find_target_cons(intf):
    # Who is consumer port? Case when not broadcast!
    # cons_rtl_port = intf.consumers[0]
    end_c = get_end_consumer(intf)
    if len(end_c) != 1:
        if len(end_c) > 1:
            log.debug(f'Find target cons: found broadcast')
        return None
    cons_rtl_port = end_c[0]
    cons_rtl_node, port_id = cons_rtl_port.node, cons_rtl_port.index
    cons_gear = cons_rtl_node.gear
    cons_port = cons_gear.in_ports[port_id]

    return cons_port


def find_target_intf(gear_name, intf_name):
    gear_mod = find(gear_name)
    rtl_node = reg['rtl/gear_node_map'][gear_mod].node

    intf_name = intf_name[1:]  # spy name always starts with _
    for i in rtl_node.local_interfaces():
        if reg['hdlgen/map'][i].basename == intf_name:
            return i


def set_waiting_edge(g, port):
    g.edge_map[port].set_color('blue')
    g.edge_map[port].set_penwidth(6)


def set_blocking_edge(g, port):
    g.edge_map[port].set_color('red')
    g.edge_map[port].set_penwidth(6)


def set_blocking_node(g, module):
    g.node_map[module].set_fillcolor('red')
    g.node_map[module].set_style('filled')


class ActivityReporter(ActivityChecker):
    def __init__(self, top, draw_graph=True, cosim_check=False):
        super().__init__(top)
        self.draw_graph = draw_graph
        self.cosim_check = cosim_check

    def after_run(self, sim):
        if self.draw_graph:
            g = graph(
                outdir=reg['results-dir'],
                node_filter=lambda g: not g.child)
        else:
            g = None

        blocking_gears = set()
        cosim_name = None
        for sim_gear in sim.sim_gears:
            if isinstance(sim_gear, SimSocket):
                cosim_name = sim_gear.gear.name
                break
        if cosim_name and self.cosim_check:
            self.cosim_activity(g, cosim_name)
        self.sim_gears_activity(g, sim, blocking_gears)

        if self.draw_graph:
            outdir = reg['results-dir']
            g.graph.write_svg(os.path.join(outdir, 'proba.svg'))

        try:
            vcd_writer = reg['VCD']
        except KeyError:
            return

        with open(os.path.join(outdir, 'issue.gtkw'), 'w') as f:
            gtkw = GTKWSave(f)
            for module in blocking_gears:
                module_sav(gtkw, module, vcd_writer.vcd_vars)

    def sim_gears_activity(self, g, sim, blocking_gears):
        for sim_gear in sim.sim_gears:
            if isinstance(sim_gear, SimSocket):
                continue
            module = sim_gear.gear
            if self.draw_graph:
                g.node_map[module].set_style('filled')
                if sim_gear not in sim.done:
                    g.node_map[module].set_fillcolor('yellow')

        def data_in_decouple(module):
            if self.draw_graph:
                set_blocking_node(g, module)
            blocking_gears.add(module)

        def not_ack(module, p):
            if self.draw_graph:
                set_blocking_edge(g, p)
            blocking_gears.add(module)

        def waiting(module, p):
            if self.draw_graph:
                set_waiting_edge(g, p)

        self.hooks['data_in_decouple'] = data_in_decouple
        self.hooks['not_ack'] = not_ack
        self.hooks['waiting'] = waiting
        super().after_run(sim)

    def cosim_activity(self, g, top_name):
        outdir = reg['results-dir']
        activity_path = os.path.join(outdir, 'activity.log')

        if not os.path.isfile(activity_path):
            return

        with open(activity_path, 'r') as log:
            for line in log:
                activity_name = line.rpartition(': ')[0]
                activity_name = activity_name.replace('top.dut.',
                                                      f'{top_name}/')
                activity_name = activity_name.rpartition('.')[0]
                activity_name = activity_name.replace('_i.', '/')
                gear_name, _, intf_name = activity_name.rpartition('/')

                # Const always has valid high
                const_regex = r'.*_const(?P<num>\d+)_s'
                const_regex_one = r'.*_const_s'
                if not (re.match(const_regex, intf_name)
                        or re.match(const_regex_one, intf_name)):
                    log.error(
                        f'Cosim spy not acknowledged: {activity_name}')

                if self.draw_graph:
                    bc_regex = r'.*_bc_(?P<num>\d+).*'
                    if re.match(bc_regex, intf_name):
                        log.debug(
                            f'Activity monitor cosim: bc not supported {activity_name}'
                        )
                        continue

                    intf = find_target_intf(gear_name, intf_name)
                    if intf is None:
                        log.error(
                            f'Cannot find matching interface for {activity_name}'
                        )
                        continue
                    if intf.is_broadcast:
                        log.debug(
                            f'Intf bc not supported {activity_name}')
                        continue

                    try:
                        prod_gear = find_target_prod(intf)
                        set_blocking_node(g, prod_gear)
                    except (KeyError, AttributeError):
                        log.debug(
                            f'Cannot find node for {activity_name}')

                    try:
                        cons_port = find_target_cons(intf)
                        set_blocking_edge(g, cons_port)
                    except (KeyError, AttributeError):
                        log.debug(
                            f'Cannot find edge for {activity_name}')



from pygears.conf import reg
from pygears.core.intf import IntfOperPlugin


async def qiter(intf):
    while True:
        data = await intf.pull()

        yield data

        intf.ack()

        if all(data.eot):
            break


class gather:
    def __init__(self, *din):
        self.din = din

    async def __aenter__(self):
        din_data = []
        for d in self.din:
            din_data.append(await d.pull())

        return tuple(din_data)

    async def __aexit__(self, exception_type, exception_value, traceback):
        for d in self.din:
            d.ack()


class PipeIntfOperPlugin(IntfOperPlugin):
    @classmethod
    def bind(cls):
        reg['gear/intf_oper/__aiter__'] = qiter



from pygears.typing import Int, Uint, is_type
from pygears.typing.visitor import TypingVisitorBase


class VGenTypeVisitor(TypingVisitorBase):
    def __init__(self, name, direction, basic_type='wire', hier=True):
        self.context = name
        self.basic_type = basic_type
        self.hier = hier
        self.direction = direction

    def visit_Int(self, type_, field, **kwds):
        return [
            f'{self.basic_type} signed [{type_.width-1}:0] {self.context}; // {type_}'
        ]

    def visit_Bool(self, type_, field, **kwds):
        return [f'{self.basic_type} [0:0] {self.context}; // {type_}']

    def visit_Uint(self, type_, field, **kwds):
        if type_.width != 0:
            return [
                f'{self.basic_type} [{type_.width-1}:0] {self.context}; // {type_}'
            ]

        return None

    visit_Ufixp = visit_Uint
    visit_Fixp = visit_Uint

    def visit_Unit(self, type_, field, **kwds):
        return None

    def visit_Union(self, type_, field, **kwds):
        res = []

        # top
        res.append(
            f'{self.basic_type} [{type_.width-1}:0] {self.context}; // {type_}')

        res.extend(
            self._complex_type_iterator([('data', type_.data),
                                         ('ctrl', type_.ctrl)]))

        return res

    def visit_Queue(self, type_, field, **kwds):
        res = []

        # top
        res.append(
            f'{self.basic_type} [{type_.width-1}:0] {self.context}; // {type_}')

        res.extend(
            self._complex_type_iterator([('data', type_.data),
                                         ('eot', type_.eot)]))
        return res

    def _complex_type_iterator(self, subt_enum):
        if not self.hier:
            return []

        res = []
        parent_context = self.context
        pos_low = 0
        for name, subt in subt_enum:
            self.context = f'{parent_context}_{name}'
            sub_wire = self.visit(subt, None)
            if sub_wire:
                res.extend(sub_wire)
                pos_high = pos_low + subt.width - 1
                if self.direction == 'input':
                    res.append(
                        f'assign {self.context} = {parent_context}[{pos_high}:{pos_low}];'
                    )
                else:
                    res.append(
                        f'assign {parent_context}[{pos_high}:{pos_low}] = {self.context};'
                    )

                pos_low = pos_high + 1

        self.context = parent_context
        return res

    def visit_Tuple(self, type_, field, **kwds):
        res = []
        res.append(
            f'{self.basic_type} [{type_.width-1}:0] {self.context}; // {type_}')

        res.extend(self._complex_type_iterator(zip(type_.fields, type_.args)))

        return res

    def visit_Array(self, type_, field, **kwds):
        if type_.data.signed:
            merge_t = Int[type_.data.width * len(type_)]
        else:
            merge_t = Uint[type_.data.width * len(type_)]

        arr_var = f'{self.context}_arr'
        res = self.visit(merge_t, type_.fields[0])
        res.append(
            f'{self.basic_type} [{type_.data.width-1}:0] {arr_var} [0:{int(len(type_))-1}];')

        high = 0
        low = 0
        for i in range(len(type_)):
            high += type_.data.width

            if self.direction == 'input':
                res.append(f'assign {arr_var}[{i}] = {self.context}[{high - 1}:{low}];')
            else:
                res.append(f'assign {self.context}[{high - 1}:{low}] = {arr_var}[{i}];')

            low += type_.data.width

        return res


def vgen_intf(dtype, name, direction, hier=True):
    valid = f'reg {name}_valid;\n'
    ready = f'reg {name}_ready;'

    if isinstance(dtype, str):
        data = f'{dtype} {name}_data;\n'
        return data + valid + ready

    if dtype.width == 0:
        data = f'wire [0:0] {name}_data;\n'
        return data + valid + ready

    vis = VGenTypeVisitor(name,
                          direction=direction,
                          basic_type='wire',
                          hier=hier)
    data = '\n'.join(vis.visit(type_=dtype, field=name)) + '\n'
    return data + valid + ready


def vgen_signal(dtype, vtype, name, direction, hier=True):
    if isinstance(dtype, str):
        return f'{dtype} {name};'

    if is_type(dtype) and dtype.width == 0:
        return f'{vtype} [0:0] {name};'

    if not hier:
        if is_type(dtype):
            width = dtype.width
            sign = 'signed' if getattr(dtype, 'signed', False) else ''
        elif isinstance(dtype, (tuple, list)):
            width = sum(d.width for d in dtype)
            sign = ''

        return f'{vtype} {sign} [{width-1}:0] {name}; // {dtype}'

    vis = VGenTypeVisitor(name,
                          basic_type=vtype,
                          direction=direction,
                          hier=hier)

    return '\n'.join(vis.visit(type_=dtype, field=name))



from pygears import gear, module, find, alternative
from pygears.sim import clk
from pygears.typing import Tuple, Uint, Union, Maybe
from .dreg import dreg

TReq = Tuple[{'addr': Uint['w_addr'], 'data': Maybe['data']}]

TWrReq = Tuple[{'addr': Uint['w_addr'], 'data': 'data'}]

TUniReq = Union[Uint['w_addr'], TWrReq]


def tdp_port0_setup(module):
    module.ram = {}


@gear(sim_setup=tdp_port0_setup)
async def tdp_port0(req, *, depth) -> b'req["data"].data':
    ram = module().ram

    async with req as (addr, (data, ctrl)):
        if ctrl:
            ram[addr] = data
        else:
            yield ram[addr]


def tdp_port1_setup(module):
    module.port0 = find('../tdp_port0')


@gear(sim_setup=tdp_port1_setup)
async def tdp_port1(req, *, depth) -> b'req["data"].data':
    ram = module().port0.ram

    async with req as (addr, (data, ctrl)):
        if ctrl:
            ram[addr] = data
        else:
            yield ram[addr]


@gear(hdl={'hierarchical': False})
def tdp(
        req0: TReq,
        req1: TReq,
        *,
        depth=b'2**w_addr',
        w_data=b'data.width',
        w_addr=b'w_addr') -> b'(req0["data"].data, req1["data"].data)':

    dout0 = req0 | tdp_port0(depth=depth)
    dout1 = req1 | tdp_port1(depth=depth)

    return dout0, dout1


@alternative(tdp)
@gear
def tdp_union(req0: TUniReq, req1: TUniReq, *, depth=b'2**w_addr'):
    wr_req_t = req0.dtype.types[1]
    req_t = TReq[wr_req_t['addr'].width, wr_req_t['data']]
    return tdp(req0 >> req_t, req1 >> req_t)



from pygears import gear, module, find
from pygears.lib import dreg
from pygears.typing import Tuple, Uint
from pygears.sim import clk

TWrDin = Tuple[{'addr': Uint['w_addr'], 'data': 'w_data'}]
TRdDin = Uint['w_addr']


@gear
async def sdp_wr_port(din, *, depth, mem) -> None:
    async with din as (addr, data):
        mem[int(addr)] = data


@gear
async def sdp_rd_port(addr, *, t, depth, mem) -> b't':
    while True:
        a = await addr.get()
        dout = mem[int(a)]
        await clk()
        yield dout


@gear(outnames=['rd_data'], hdl={'hierarchical': False}, enablement=b'latency not in [0, 2]')
def sdp(wr_addr_data: TWrDin,
        rd_addr: TRdDin,
        *,
        depth=b'2**w_addr',
        w_data=b'w_data',
        w_addr=b'w_addr',
        latency=1,
        mem=None) -> b'w_data':
    """Short for Simple Dual-Port RAM. Supports simultaneous read and write
    operations i.e. ``rd_addr`` interface reads from the RAM while the
    ``wr_addr_data`` interface writes to it. It has a sigle output interface
    for read data called ``rd_data``. Since the memory doesn't use the DTI,
    adapters are needed to convert the DTI protocol to the memory interface.
    The gear consists of three submodules implemented in SystemVerilog.

    Write port adapter for the SDP RAM is called ``sdp_wr_port``. It translates
    the DTI interface to the appropriate SDP signals. The ``data`` and ``addr``
    fields of the :class:`Tuple` type are passed to the `addr` and `data` inputs
    of the memory while `valid` is used as enable input to the memory.

    Read port adapter for the SDP RAM is called ``sdp_rd_port``. It translates
    the DTI interface to the appropriate SDP signals. Data from the input
    interface is used as an address for the memory and the read data is passed
    to the output interface. Control logic is implemented for correct generation
    of all `ready` and `valid` signals for the input and output interfaces.

    The third submodule called ``sdp_mem`` is the memory itself.

    Args:
        w_data: Width of the data bus
        w_addr: Width of the address bus
        depth: Depth of the memory

    Returns:
        Data read from the memory. Same type as the ``data`` field of the
          ``TWrDin`` :class:`Tuple` i.e. the data beeing writen to memory.
    """

    if mem is None:
        mem = {}

    wr_addr_data | sdp_wr_port(depth=depth, mem=mem)
    return rd_addr | sdp_rd_port(t=wr_addr_data.dtype['data'], depth=depth, mem=mem)



from pygears import gear, module, alternative
from pygears.typing import Union, Uint
from pygears.lib.ccat import ccat


def demux_type(dtypes, mapping):
    tout = [None] * (max(mapping.values()) + 1)
    for idin, idout in mapping.items():
        din_t = dtypes.types[idin]
        if tout[idout] is not None:
            if tout[idout] != din_t:
                tout[idout] = Uint[max(din_t.width, tout[idout].width)]
        else:
            tout[idout] = din_t

    assert not any(t is None for t in tout)

    return tuple(tout)


def full_mapping(dtypes, mapping, use_dflt):
    if mapping is None:
        mapping = dflt_map(dtypes)

    if not use_dflt:
        return mapping

    dout_num = max(mapping.values()) + 1
    fm = mapping.copy()

    for i in range(len(dtypes.types)):
        if i not in mapping:
            fm[i] = dout_num

    return fm


def dflt_map(dtypes):
    return {i: i for i in range(len(dtypes.types))}


@gear
async def demux(
    din: Union,
    *,
    use_dflt=True,
    mapping=b'dflt_map(din)',
    _full_mapping=b'full_mapping(din, mapping, use_dflt)',
) -> b'demux_type(din, _full_mapping)':
    async with din as (data, ctrl):
        tout = module().tout
        if not isinstance(tout, tuple):
            yield tout.decode(int(data))
        else:
            dout = [None] * len(module().tout)
            ctrl = _full_mapping[int(ctrl)]
            dout[ctrl] = tout[int(ctrl)].decode(int(data))

            yield tuple(dout)


@gear
def demux_ctrl(din: Union, *, use_dflt=True, mapping=None):
    fdemux = demux
    if mapping is None:
        fdemux = demux
    else:
        fdemux = fdemux(mapping=mapping, use_dflt=use_dflt)

    dout = din | fdemux
    if not isinstance(dout, tuple):
        dout = (dout, )

    return (din[1], *dout)


@alternative(demux)
@gear
def demux_by(ctrl: Uint, din, *, fcat=ccat, nout=None, use_dflt=True, mapping=None):
    if nout is None:
        nout = 2**ctrl.dtype.width

    demux_din = fcat(din, ctrl) \
        | Union[(din.dtype, ) * nout] \

    if mapping is None:
        return demux_din | demux
    else:
        return demux_din | demux(mapping=mapping, use_dflt=use_dflt)


@gear
def demux_zip(din: Union) -> b'(din[1], ) + tuple(t for t in din.types)':
    pass



import os
from enum import IntEnum

from traceback import extract_stack, extract_tb

from pygears.definitions import ROOT_DIR

from .pdb_patch import patch_pdb, unpatch_pdb
from .registry import PluginBase, reg


class TraceLevel(IntEnum):
    debug = 0
    user = 1


def set_trace_level(var, val):
    var._val = val
    if val == TraceLevel.user:
        patch_pdb()
    else:
        unpatch_pdb()


def parse_trace(s):
    if reg['trace/level'] == TraceLevel.debug:
        return s
    else:
        trace_fn = os.path.abspath(s.filename)
        is_internal = any(trace_fn.startswith(d) for d in reg['trace/ignore'])

        is_decorator_gen = '<decorator-gen' in trace_fn

        if is_internal or is_decorator_gen:
            return None

        return s


def enum_formated_stack_frames(summary):
    for frame, display in zip(summary, summary.format()):
        if parse_trace(frame):
            yield display


def enum_traceback(tr):
    yield from enum_formated_stack_frames(extract_tb(tr))


def enum_stacktrace():
    yield from enum_formated_stack_frames(extract_stack())


class TraceFormatPlugin(PluginBase):
    @classmethod
    def bind(cls):
        reg['trace/hooks'] = []

        reg.confdef('trace/level', setter=set_trace_level, default=TraceLevel.user)

        reg.confdef(
            'trace/ignore',
            default=[
                os.path.join(ROOT_DIR, d) for d in ['core', 'conf', 'sim', 'hls', 'hdl']
            ] + [os.path.dirname(os.__file__)])



from pygears import gear
from pygears.typing import Uint, code


@gear
async def iceil(din: Uint['T'], *, div=4) -> Uint['T']:
    """Performs division of the input value and return the ceiling of the
    calculated value

    Args:
        div: The divisor value
    """
    async with din as val:
        yield code((val + div - 1) // div, din.dtype)



from pygears import bind
from pygears.lib.decouple import decouple_din
from pygears.lib import const
from pygears.sim import log
from pygears.conf import inject, Inject
from pygears.core.gear import Gear
from pygears.core.graph import get_producer_queue, get_source_producer


class ActivityChecker:
    @inject
    def __init__(self, top, sim=Inject('sim/simulator')):
        sim.events['before_run'].append(self.before_run)
        sim.events['after_run'].append(self.after_run)
        reg['sim/activity'] = self
        self.blockers = {}
        self.handshakes = set()
        self.hooks = {}

    def intf_pull_start(self, intf):
        consumer = intf.producer
        producer = intf.in_queue.intf.consumers[0]
        self.blockers[consumer] = producer
        return True

    def intf_pull_done(self, intf):
        consumer = intf.producer
        self.handshakes.add(consumer)
        self.handshakes.add(self.blockers[consumer])
        del self.blockers[consumer]
        return True

    def before_timestep(self):
        self.handshakes.clear()

    @inject
    def before_run(self, sim, sim_map=Inject('sim/map')):
        for module, sim_gear in sim_map.items():
            if isinstance(module, Gear):
                ports = module.in_ports
            else:
                continue

            for p in ports:
                p.consumer.events['pull_start'].append(self.intf_pull_start)
                p.consumer.events['pull_done'].append(self.intf_pull_done)

    def get_port_status(self, port):
        q = get_producer_queue(port)

        if q._unfinished_tasks:
            return "active"

        prod_port = get_source_producer(port, sim=True).consumers[0]
        if prod_port in self.handshakes:
            return "handshaked"

        if prod_port in self.blockers.values():
            return "waited"

        return "empty"

    def after_run(self, sim):
        for sim_gear in sim.sim_gears:
            module = sim_gear.gear

            if module.definition == decouple_din:
                if not module.queue.empty():
                    if 'data_in_decouple' in self.hooks:
                        self.hooks['data_in_decouple'](module)
                    log.error(f'Data left in decouple: {module.name}')

            for p in module.in_ports:
                status = self.get_port_status(p)

                if status == "active":
                    src_port = get_source_producer(p, sim=True).consumers[0]

                    if src_port.gear.definition == const:
                        # Skip constants since they are never done
                        continue

                    if 'not_ack' in self.hooks:
                        self.hooks['not_ack'](module, p)
                    log.error(
                        f'{src_port.gear.name}.{src_port.basename} -> {module.name}.{p.basename} was not acknowledged'
                    )

                if status == "waited":
                    src_port = self.blockers[p]
                    if 'waiting' in self.hooks:
                        self.hooks['waiting'](module, p)
                    log.debug(
                        f'{p.gear.name}.{p.basename} waiting on {src_port.gear.name}.{src_port.basename}'
                    )



from pygears import gear, Intf
from pygears.core.gear_inst import find_current_gear_frame
from pygears.core.intf import IntfOperPlugin
from pygears.lib.union import field_sel
from pygears import module, reg


@gear
async def sieve(din, *, key) -> b'din[key]':
    """Outputs a slice of the ``din`` input interface. Can be instantiated with
    the slicing statement: ``din[key]``.

    Args:
        key: A single key or a sequence of keys with which to slice the input
          interface.

    Returns:
        A sliced interface

    Which keys are exactly supported depends on the type of the ``din`` input
    interface, so checkout the __getitem__ method of the specific type. If for
    an example we have an interface of the type :class:`Uint[8] <Uint>` ::

        din = Intf(Uint[8])

    we could slice it using Python index operator to obtain a high nibble:

    >>> din[4:]
    Intf(Uint[4])

    which outputs an interface of the type :class:`Uint[4] <Uint>`. The same
    would be achieved if the ``sieve`` gear were instantiated explicitly:

    >>> sieve(din, key=slice(4, None, None))
    Intf(Uint[4])
    """

    async with din as d:
        dout = []
        for i in key:
            dout.append(d[i])

        if len(key) == 1:
            dout = dout[0]

        yield module().tout(dout)


def get_obj_var_name(frame, obj):
    for var_name, var_obj in frame.f_locals.items():
        if obj is var_obj:
            return var_name
    else:
        return None


def maybe_obtain_intf_var_name(intf):
    frame = find_current_gear_frame()
    if frame is None:
        return None

    return get_obj_var_name(frame, intf)


def get_sieve(self, index):
    naming = reg['gear/naming/pretty_sieve']
    norm_index = self.dtype.index_norm(index)

    # Try to obtain variable to which interface was assigned to form a better
    # name for the sieve
    name = 'sieve'
    if naming:
        name = maybe_obtain_intf_var_name(self)
        if name is None:
            name = 'sieve'

    if not isinstance(index, tuple):
        index = (index, )

    name_appendices = []
    for ind_id, ind in enumerate(norm_index):
        if isinstance(ind, slice):
            name_appendices.append(f'{ind.start}v{ind.stop}')
        else:
            try:
                # Try to obtain original index name to form a better name
                # for the sieve
                orig_ind = str(index[ind_id])

                if orig_ind.isalnum():
                    ind = orig_ind

            except IndexError:
                pass

            name_appendices.append(f'{ind}')

    return self | sieve(key=norm_index,
                        name='_'.join([name] + name_appendices))


def get_select(self, index):
    return field_sel(index, self)


def getitem(self, index):
    if isinstance(index, Intf):
        return get_select(self, index)
    else:
        return get_sieve(self, index)


class GetitemIntfOperPlugin(IntfOperPlugin):
    @classmethod
    def bind(cls):
        reg['gear/intf_oper/__getitem__'] = getitem
        reg['gear/naming/pretty_sieve'] = False



import inspect
import pprint
import textwrap

from pygears import GearDone, gear, datagear
from pygears.lib import decouple
from pygears.util.utils import quiter, gather
from pygears.typing import Uint, Any, typeof, TypeMatchError
from pygears.sim import sim_assert, log, clk
from pygears.typing.visitor import TypingVisitorBase


class TypingYieldVisitorBase:
    def visit(self, data, dtype):
        visit_func_name = f'visit_{dtype.__name__.lower()}'

        visit_func = getattr(self, visit_func_name, self.visit_default)
        ret = visit_func(data, dtype)
        if inspect.isgenerator(ret):
            yield from ret
        else:
            yield ret

    def visit_default(self, data, dtype):
        yield data


class TypeDrvVisitor(TypingYieldVisitorBase):
    def visit_queue(self, data, dtype):
        for d, eot in quiter(data):
            if type(d) == dtype:
                yield d
            else:
                for ret in self.visit(d, dtype.sub()):
                    if dtype.lvl == 1:
                        yield (ret, Uint[1](eot))
                    else:
                        yield (ret[0], eot @ ret[1])


def typeseq(t, v):
    if type(v) == t:
        yield v
    else:
        try:
            for d in TypeDrvVisitor().visit(v, t):
                try:
                    yield t(d)
                except (TypeError, ValueError) as e:
                    log.error(f'{e} - Cannot convert value "{d}" to type "{repr(t)}"')
        except (TypeError, ValueError):
            log.error(f'Cannot convert sequence "{v}" to the "{repr(t)}"')


@gear
async def drv(*, t, seq) -> b't':
    """Outputs one data at the time from the iterable ``seq`` cast to the type
    ``t``.

    Args:
        t: Type of the data to output
        seq: An iterable generating data to be output

    Returns:
        Data of the type ``t``

    >>> drv(t=Uint[8], seq=range(10))

    If ``t`` is a :class:`Queue` type of certain level, then ``seq`` should
    generate nested iterables of the same level::

        q1 = ((11, 12), (21, 22, 23))
        q2 = ((11, 12, 13))

    >>> drv(t=Queue[Uint[8], 2], seq=[q1, q2])
    """

    for s in seq:
        for val in typeseq(t, s):
            yield val

    raise GearDone


@gear
async def secdrv(seqin, *, t) -> b't':
    async with seqin as seq:
        for val in seq:
            if type(val) == t:
                yield val
            else:
                for d in TypeDrvVisitor().visit(val, t):
                    yield t(d)


class Partial:
    def __new__(cls, val):
        if isinstance(val, Partial):
            return val
        else:
            obj = super().__new__(cls)
            obj.__init__(val)
            return obj

    def __init__(self, val):
        self._val = val

    @property
    def val(self):
        return self._val


class TypeMonitorVisitor:
    def __init__(self, dtype):
        self.data = None
        self.dtype = dtype

    def __bool__(self):
        return isinstance(self.data, Partial)

    def append(self, elem):
        self.data = self.visit(self.data, elem, self.dtype)
        return self.data

    def visit(self, data, elem, dtype):
        visit_func_name = f'visit_{dtype.__name__.lower()}'

        visit_func = getattr(self, visit_func_name, self.visit_default)

        return visit_func(data, elem, dtype)

    def visit_default(self, data, elem, dtype):
        return elem

    def visit_queue(self, data, elem, dtype):
        if dtype.lvl == 1:
            sub_elem = elem[0]
        else:
            sub_elem = elem.sub()

        if not data:
            sub_data = None
            data = []
        else:
            data = data.val
            if isinstance(data[-1], Partial):
                sub_data = data.pop()
            else:
                sub_data = None

        sub_data = self.visit(sub_data, sub_elem, dtype.sub())
        data.append(sub_data)

        eot = all(elem.eot)
        if eot and (not isinstance(sub_data, Partial)):
            return data
        else:
            return Partial(data)


@gear
async def mon(din, *, t=b'din') -> Any:
    v = TypeMonitorVisitor(t)
    data = None
    while (isinstance(data, Partial) or data is None):
        # print('Monitor waiting')
        item = await din.get()
        # print('Monitor got: ', item)
        data = v.visit(data, item, t)

    # print('Monitor emits: ', data)
    yield data


# TODO: What if the dut quits earlier, catch that situation
@gear
async def scoreboard(*din: b't', report=None, cmp=None) -> None:
    """Generic scoreboard used from comparing actual results from the DUT to
    expected results. Eventual mismatches are asserted using the ``sim_assert``
    function meaning that any ``error`` behaviour is controled via the ``sim``
    logger ``error`` settings.

    Args:
        din: Outputs from both the DUT and ref. model. All intpus are a PyGears
          interface
        report: List to with comparison results are appended
        tolerance: Optional tolerance when comparing results. The DUT result must
          be in the (expected - tolerance, expected+tolerance) range

    Returns:
        None
    """
    def match_exact(x, y):
        return x == y

    if cmp is None:
        cmp = match_exact

    # print(f'Number of matches: -\r', end='')
    cnt = 0
    match_cnt = 0
    try:
        while True:
            items = []
            for d in din:
                items.append(await d.get())

            match = all(cmp(v, items[0]) for v in items)

            if report is not None:
                report.append({'match': match, 'items': items})

            cnt += 1
            if match:
                match_cnt += 1

            sim_assert(match, f'mismatch on #{cnt}: {items[0]}, {items[1]}')

            # print(f'Number of matches: {match_cnt:>4}/{cnt:>4}\r', end='')

    except GearDone as e:
        log.info(f'Number of matches: {match_cnt:>4}/{cnt:>4}')
        offending = None

        if len(items):
            offending = len(items)

        if any(not d.empty() for d in din):
            for i, d in enumerate(din):
                if d.empty():
                    offending = i
                    break

        if offending is not None:
            log.error(f"Input {offending} didn't produce enough output data")

        raise e


class ErrorVisitor:
    def visit(self, a, b):
        # if not typeof(type(b), type(a)):
        # if type(a) != type(b):
        #     breakpoint()
        #     raise ValueError(f'{type(b)} cannot be matched to {type(a)}')

        type_ = type(a)
        for c in type_.mro():
            visit_func_name = c.__name__

            if hasattr(self, visit_func_name):
                return getattr(self, visit_func_name)(a, b)

        else:
            return self.visit_default(a, b)

    def visit_default(self, a, b):
        if a != b:
            raise ValueError(f'Got: {a}, expected: {b}')

    def Queue(self, a, b):
        if a.eot != b.eot:
            raise ValueError(f'Got Queue eot: {a.eot}, expected: {b.eot}')

        err = None
        try:
            self.visit(a.data, b.data)
        except ValueError as e:
            err = e

        if err:
            raise ValueError(f'{str(err)}\n - when matching {a} to {b}')

    def Maybe(self, a, b):
        if a.ctrl != b.ctrl:
            if a.ctrl:
                raise ValueError(f'Got: SOME, expected: NONE')
            else:
                raise ValueError(f'Got: NONE, expected: SOME')

        if not a.ctrl:
            return

        va = type(a).types[a.ctrl].decode(a.data)
        vb = type(b).types[b.ctrl].decode(b.data)

        self.visit(va, vb)


    def Array(self, a, b):
        for i in range(len(type(a))):
            err = None
            try:
                self.visit(a[i], b[i])
            except ValueError as e:
                err = e

            if err:
                raise ValueError(f'{str(err)}\n - when matching element {i} of {a} to {b}')

    def Tuple(self, a, b):
        for i in range(len(type(a))):
            err = None
            try:
                self.visit(a[i], b[i])
            except ValueError as e:
                err = e

            if err:
                raise ValueError(f'{str(err)}\n - when matching element {i} of {a} to {b}')


@gear
async def check(din, *, ref, cmp=None):
    """Checks equality of input data with expected.

    Args:
        ref: A list of expected values

    Returns:
        None

    If type ``din`` is a :class:`Queue` of certain level, then ``ref`` should
    generate nested iterables of the same level
    """
    def match_exact(x, y):
        return x == y

    if cmp is None:
        cmp = match_exact

    iter_ref = iter(ref)
    ref_seq = iter(())

    try:
        items = []
        while True:
            data = await din.get()
            items.append(data)

            try:
                ref_item = next(ref_seq)
            except StopIteration:
                ref_seq = typeseq(din.dtype, next(iter_ref))
                ref_item = next(ref_seq)

            cmp_res = cmp(data, ref_item)

            if not cmp_res:
                if cmp is match_exact:
                    v = ErrorVisitor()
                    err = None
                    try:
                        v.visit(data, ref_item)
                    except ValueError as e:
                        err = e

                    if err:
                        log.error(f'mismatch in item #{len(items)}\n {str(err)}')
                    else:
                        breakpoint()
                else:
                    log.error(f'mismatch in item #{len(items)}. Got: {data}, expected: {ref_item}')

    except GearDone:
        ref_empty = False
        try:
            next(ref_seq)
        except StopIteration:
            try:
                next(iter_ref)
            except StopIteration:
                ref_empty = True

        if ref_empty:
            log.info(f'Number of matches: {len(items)}')
        else:
            log.error(f"mismatch in number of items, got '{len(items)}' but expected '{len(ref)}'. "
                      f"\ngot:\n{textwrap.indent(pprint.pformat(items), ' '*4)}"
                      f"\nexp:\n{textwrap.indent(pprint.pformat(ref), ' '*4)}")

    except (GearDone, StopIteration):
        log.error(f"mismatch in number of items {len(items)} vs {len(ref)}. "
                  f"\ngot:\n{textwrap.indent(pprint.pformat(items), ' '*4)}"
                  f"\nexp:\n{textwrap.indent(pprint.pformat(ref), ' '*4)}")

    raise GearDone


def tlm_verif(*seq, f, ref):
    res_tlm = tuple(s | drv for s in seq) \
        | f \
        | mon

    ref_tlm = seq | ref

    report = []
    scoreboard(res_tlm, ref_tlm, report=report)

    return report


def verif(*stim, f, ref, delays=None, cmp=None, check_timing=False, make_report=False):
    """Verification environment for comparing DUV results with reference model.
    The environment instantiates the DUV and reference model and drives the
    passed stimulus to both. The outpus are passed to the scoreboard which
    compares the results. Outputs are decoupled to ensure there is no connection
    between the DUV and the environment. Optional delays can be added to all
    input and output interfaces.

    Args:
        stim: Input stimulus
        f: Gear to be verified
        ref: Gear used as reference model
        delays: List of delays for all inputs and outputs
        tolerance: Tolerance window when performing equality checks

    Returns:
        A report dictionary with pass/fail statistics
    """

    if stim:
        res_tlm = stim | f
        ref_tlm = stim | ref
    else:
        res_tlm = f
        ref_tlm = ref

    if check_timing:
        ref_tlm, res_tlm = match_timing(ref_tlm, res_tlm)

    if not isinstance(res_tlm, tuple):
        res_tlm = (res_tlm, )
        ref_tlm = (ref_tlm, )

    if make_report:
        report = [[] for _ in range(len(res_tlm))]
    else:
        report = [None for _ in range(len(res_tlm))]

    if delays is None:
        delays = (None, ) * len(res_tlm)

    assert len(ref_tlm) == len(res_tlm)
    assert len(delays) == len(res_tlm)

    for r, res_intf, ref_intf, d in zip(report, res_tlm, ref_tlm, delays):
        if d is not None:
            res_intf = res_intf | d

        res_intf = res_intf | decouple(depth=0)
        ref_intf = ref_intf | decouple(depth=0)

        scoreboard(res_intf, ref_intf, report=r, cmp=cmp)

    return report


def directed(*stim, f, ref, delays=None, cmp=None):
    """Similar to ``verif`` function, except ``ref`` is a list of expected results"""
    if stim:
        res = stim | f
    else:
        res = f

    if res is None:
        raise Exception(f'Module under test has no output interface')

    if not isinstance(res, tuple):
        res = (res, )
        ref = (ref, )

    if delays is None:
        delays = (None, ) * len(res)

    assert len(ref) == len(res)
    assert len(delays) == len(res)

    for ref_inst, res_inst, delay_inst in zip(ref, res, delays):
        if delay_inst is not None:
            res_inst = res_inst | delay_inst
        res_inst | check(ref=ref_inst, cmp=cmp)


def directed_on_the_fly(*stim, f, refs, delays=None):
    """Similar to ``directed`` function, except ``ref`` is a generator and
    checking is done `on-the-fly`"""
    res_tlm = stim | f

    if not isinstance(res_tlm, tuple):
        res_tlm = (res_tlm, )

    if delays is None:
        delays = (None, ) * len(res_tlm)

    report = [[] for _ in range(len(res_tlm))]
    for r, res_intf, ref, d in zip(report, res_tlm, refs, delays):
        if d is not None:
            res_intf = res_intf | d

        res_intf = res_intf | decouple(depth=0)
        # ref = ref | decouple(depth=0)

        scoreboard(res_intf, ref, report=r)

    return report


@gear
async def match_timing(*din) -> b'din':
    if any(not d.empty() for d in din):
        all_valid = all(not d.empty() for d in din)
        sim_assert(all_valid, f'not all inputs valid at the same time')

        async with gather(*din) as data:
            yield data

    await clk()


@datagear
def collect(val, *, result):
    result.append(val)



import shutil
import os
from pygears import Intf, find, reg
from pygears.util.fileio import expand
from .common import list_hdl_files
from .generate import generate as hdlgen_generate


def hdlgen(top=None,
           lang=None,
           toplang=None,
           copy_files=False,
           generate=True,
           outdir=None,
           **conf):

    if lang is None:
        lang = reg['hdl/lang']
    else:
        # TODO: should we save/restore previous setting for 'hdl/lang'?
        reg['hdl/lang'] = lang

    if outdir is None:
        outdir = reg['results-dir']

    conf['outdir'] = expand(outdir)
    outdir = conf['outdir']

    if isinstance(top, tuple):
        top = top[0]

    if isinstance(top, Intf):
        top = top.producer.gear

    if top is None:
        top = reg['gear/root']
    elif isinstance(top, str):
        mtop = find(top)
        if mtop is None:
            raise Exception(f'No module "{top}" found')

        top = mtop
    else:
        top = top

    if toplang is None:
        toplang = reg['hdl/toplang']
        if toplang is None:
            toplang = reg['hdl/lang']
    else:
        reg['hdl/top'] = top

    reg['hdl/toplang'] = toplang

    reg['svgen/conf'] = conf
    for oper in reg[f'{lang}gen/flow']:
        oper(top, conf)

    os.makedirs(outdir, exist_ok=True)
    if generate:
        hdlgen_generate(top, conf)

        for (modname, lang), (fn, fn_dis) in reg['hdlgen/disambig'].items():
            with open(fn) as fin:
                with open(fn_dis, 'w') as fout:
                    mod = fin.read()
                    mod = mod.replace(f'module {modname}',
                                      f'module {modname}_{lang}')
                    fout.write(mod)

    if copy_files and generate:
        for fn in list_hdl_files(top.name, outdir=outdir, rtl_only=True):
            modname, ext = os.path.splitext(os.path.basename(fn))
            if (modname, ext[1:]) in reg['hdlgen/disambig']:
                continue

            try:
                shutil.copy(fn, outdir)
            except shutil.SameFileError:
                pass
            except FileNotFoundError:
                pass

    return top



import inspect
import sys
import json
import os
import pprint
import runpy

from .log import conf_log
from .registry import PluginBase, reg
from .utils import dict_generator


def print_registry():
    # monkey patch sorting
    pprint._safe_key.__lt__ = lambda x, y: True

    preg = pprint.pformat(reg)
    conf_log().info(f'Registry settings:\n{preg}')


def load_rc_from_dir(rc_fn, dirname):
    rc_path = os.path.join(dirname, f'{rc_fn}.py')
    if os.path.exists(rc_path):
        runpy.run_path(rc_path)
        return

    conf = None
    rc_path = os.path.join(dirname, f'{rc_fn}.yaml')
    if os.path.exists(rc_path):
        with open(rc_path) as f:
            try:
                import yaml
                conf = yaml.safe_load(f)
            except ImportError:
                conf_log().warning(
                    f'PyGears YAML configuration file found'
                    f'at "{rc_path}", but yaml python package not '
                    f'installed')

    rc_path = os.path.join(dirname, f'{rc_fn}.json')
    if os.path.exists(rc_path):
        with open(rc_path) as f:
            conf = json.load(f)

    if conf:
        for c_list in dict_generator(conf):
            keys = '/'.join([str(x) for x in c_list[:-1]])
            reg[keys] = c_list[-1]


def load_rc(rc_fn, dirname=None):
    search_dirs = []

    if dirname is None:
        if hasattr(sys.modules['__main__'], '__file__'):
            dirname = os.path.dirname(sys.modules['__main__'].__file__)
        else:
            dirname = os.getcwd()

    while dirname != os.path.abspath(os.sep):
        search_dirs.append(dirname)
        dirname = os.path.abspath(os.path.join(dirname, '..'))

    home_path = os.path.expanduser("~")
    if home_path not in search_dirs:
        search_dirs.append(home_path)

    pg_home = os.path.join(home_path, '.pygears')
    if pg_home not in search_dirs:
        search_dirs.append(pg_home)

    unique_search_dirs = list(dict.fromkeys(search_dirs))
    for path in reversed(unique_search_dirs):
        load_rc_from_dir(rc_fn, path)



from pygears.lib import drv, collect, mon
import inspect
from pygears.sim import sim
from pygears.typing import Integer, Int, Uint, is_type, typeof, Any
from pygears.core.partial import extract_arg_kwds, combine_arg_kwds, all_args_specified
from pygears.core.gear_inst import resolve_args
from pygears import reset, clear
from pygears.sim import cosim


def infer_dtype(val, dtype):
    if is_type(type(val)):
        return type(val)

    if not is_type(dtype) or typeof(dtype, Any):
        if isinstance(val, int):
            if val < 0:
                return type(Int(val))
            else:
                return type(Uint(val))

    if dtype.specified:
        return dtype

    return type(dtype.base(val))


def call(f, *args, **kwds):
    clear()
    kwd_intfs, kwd_params = extract_arg_kwds(kwds, f)
    args_comb = combine_arg_kwds(args, kwd_intfs, f)

    paramspec = inspect.getfullargspec(f.func)
    args, annotations = resolve_args(args_comb, paramspec.args,
                                     paramspec.annotations, paramspec.varargs)

    dtypes = [infer_dtype(args[arg], annotations[arg]) for arg in args]

    seqs = [drv(t=t, seq=[v]) for t, v in zip(dtypes, args_comb)]

    outputs = f(*seqs, **kwd_params)

    if isinstance(outputs, tuple):
        res = [[] for _ in outputs]

        for o, r in zip(outputs, res):
            collect(o | mon, result=r)
    else:
        res = [[]]
        collect(outputs | mon, result=res[0])

    sim(check_activity=False)

    if isinstance(outputs, tuple):
        return res
    else:
        return res[0]



from pygears import gear
from pygears.core.partial import Partial
from pygears.typing import Number, Queue, saturate


@gear
async def reduce(din: Queue, init, *, f) -> b'init':
    acc: init.dtype

    async with init as i:
        acc = i
        async for (d, eot) in din:
            if isinstance(f, Partial):
                async with f(acc, d) as fout:
                    acc = fout

                    if eot:
                        yield acc
            else:
                acc = f(acc, d)

                if eot:
                    yield acc


@gear
def accum(din: Queue[Number], init: Number, *, cast=saturate) -> b'init':
    def add(x, y):
        return cast(x + y, t=init.dtype)

    return reduce(din, init, f=add)


@gear
def qmax(din: Queue[Number]) -> b'din.data':
    def fmax(x, y):
        return x if x > y else y

    return reduce(din, din.dtype.data.min, f=fmax)
